from __future__ import annotations

import time
import random
import urllib.parse
import typing as t
from lzl.types import BaseModel, PrivateAttr, eproperty
from pydantic.networks import AnyHttpUrl
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    Protocol,
    Awaitable,
    cast,
)
from .errors import (
    ConnectError,
    MaxAttemptsError,
    MaxRedirectsError,
    UnexpectedResponse,
)
from . import logging

# import requests
from lzl.api import aioreq
from .utils import parse_host


T = TypeVar("T")


if t.TYPE_CHECKING:
    from lzl.api.aioreq import Response, AsyncResponse

class ConnURL(BaseModel):
    host: AnyHttpUrl
    _extra: Dict[str, Any] = PrivateAttr(default_factory = dict)

    @eproperty
    def safe_url(self) -> str:
        """
        Returns the url
        """
        base = f'{self.host.scheme}://'
        if self.host.username and self.host.password:
            base += f'{self.host.username}:****@'
        elif self.host.password:
            base += ':****@'
        base += self.host.host
        if self.host.port:
            base += f':{self.host.port}'
        return base

    @eproperty
    def url(self) -> str:
        """
        Returns the url
        """
        base = f'{self.host.scheme}://'
        if self.host.username and self.host.password:
            base += f'{self.host.username}:{self.host.password}@'
        elif self.host.password:
            base += f':{self.host.password}@'
        base += self.host.host
        if self.host.port:
            base += f':{self.host.port}'
        return base

    @eproperty
    def bare(self) -> str:
        """
        Returns the url
        """
        base = f'{self.host.scheme}://{self.host.host}'
        if self.host.port:
            base += f':{self.host.port}'
        return base


    def with_uri(self, uri: str) -> str:
        """
        Returns the url
        """
        return urllib.parse.urljoin(self.url, uri)

class ConnectionClient:
    def __init__(
        self, 
        hosts: List[ConnURL],
        log_config: logging.LogConfig,
        timeout: int = 5,
        max_redirects: int = 2,
        max_attempts_per_host: int = 2,
        
    ):
        self.hosts: List[ConnURL] = hosts
        self.log_config = log_config
        self.timeout = timeout
        self.max_redirects = max_redirects
        self.max_attempts_per_host = max_attempts_per_host
        from lzl.api.aioreq import Client, exceptions
        self._excs = exceptions
        self.io = Client(
            timeout = timeout,
            retries = max_attempts_per_host,
        )
        self.timeout = timeout

    def fetch_response_with_host(
        self,
        host: ConnURL,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: bool = False,
    ) -> 'Response':
        """Fetches a response from a particular host, and returns the status
        code and headers. The response body is written to the given destination.

        This will follow up to the maximum number of redirects.

        Args:
            host (str): The host to fetch the response from.
            port (int): The port to use when connecting to the host.
            method (str): The HTTP method to use.
            uri (str): The URI to fetch.
            json (dict): The JSON to send in the request body.
            headers (dict): The headers to send in the request in addition to the content type.
                The dictionary should be accessed only using lowercase keys.

        Returns:
            requests.Response: The response from the server.

        Raises:
            ConnectTimeout: If the connection times out.
            MaxRedirectsError: If the maximum number of redirects is exceeded.
            UnexpectedResponse: If the server returns a response we didn't expect.
        """
        redirect_path = []
        original_host = host.with_uri(uri)
        current_host = original_host
        while len(redirect_path) < self.max_redirects:
            try:
                response = self.io.request(
                    method,
                    current_host,
                    json = json,
                    headers = headers,
                    timeout = self.timeout,
                    allow_redirects = False,
                    stream = stream,
                )
                if response.is_redirect:
                    redirected_to = response.headers["Location"]
                    if stream: response.close()
                    current_host = redirected_to
                    redirect_path.append(redirected_to)
                    continue

                if response.status_code < 200 or response.status_code > 299:
                    raise UnexpectedResponse(
                        current_host,
                        f"Unexpected response from {current_host}: {response.status_code} {response.reason}",
                    )

                return response
            except self._excs.ConnectTimeout as e:
                raise ConnectError(
                    f"Connection to {host.safe_url} timed out", current_host
                ) from e
            except self._excs.ConnectionError as e:
                raise ConnectError(
                    f"Connection to {host.safe_url} was refused", current_host
                ) from e
        raise MaxRedirectsError(original_host, redirect_path)


    async def afetch_response_with_host(
        self,
        host: ConnURL,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: bool = False,
    ) -> 'Response':
        """Fetches a response from a particular host, and returns the status
        code and headers. The response body is written to the given destination.

        This will follow up to the maximum number of redirects.

        Args:
            host (str): The host to fetch the response from.
            port (int): The port to use when connecting to the host.
            method (str): The HTTP method to use.
            uri (str): The URI to fetch.
            json (dict): The JSON to send in the request body.
            headers (dict): The headers to send in the request in addition to the content type.
                The dictionary should be accessed only using lowercase keys.

        Returns:
            requests.Response: The response from the server.

        Raises:
            ConnectTimeout: If the connection times out.
            MaxRedirectsError: If the maximum number of redirects is exceeded.
            UnexpectedResponse: If the server returns a response we didn't expect.
        """
        redirect_path = []
        original_host = host.with_uri(uri)
        current_host = original_host
        while len(redirect_path) < self.max_redirects:
            try:
                response: 'Response' = await self.io.arequest(
                    method,
                    current_host,
                    json = json,
                    headers = headers,
                    timeout = self.timeout,
                    allow_redirects = False,
                    stream = stream,
                )
                if response.is_redirect:
                    redirected_to = response.headers["Location"]
                    if stream: response.close()
                    current_host = redirected_to
                    redirect_path.append(redirected_to)
                    continue

                if response.status_code < 200 or response.status_code > 299:
                    raise UnexpectedResponse(
                        current_host,
                        f"Unexpected response from {current_host}: {response.status_code} {response.reason}",
                    )

                return response
            except self._excs.ConnectTimeout as e:
                raise ConnectError(
                    f"Connection to {host.safe_url} timed out", current_host
                ) from e
            except self._excs.ConnectionError as e:
                raise ConnectError(
                    f"Connection to {host.safe_url} was refused", current_host
                ) from e
        raise MaxRedirectsError(original_host, redirect_path)
    
    def fetch_response(
        self,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: bool = False,
        initial_host: Optional[ConnURL] = None,
        query_info: Optional[logging.QueryInfoLazy] = None,
    ) -> 'Response':
        """Fetches a response from the server by requesting it from a random node. If
        a connection error occurs, this method will retry the request on a different
        node until it succeeds or the maximum number of attempts is reached.

        Args:
            method (str): The HTTP method to use for the request.
            uri (str): The URI of the request.
            json (dict): The json data to send with the request.
            headers (dict): The headers to send with the request.
            stream (bool): If True, the response will be streamed.
            initial_host (Optional[ConnURL]): The host to try first. If None,
                a random host will be chosen.
            query_info (Optional[logging.QueryInfo]): The query info for slow
                query logging, or None to disable slow query logging regardless of
                the log configuration.

        Returns:
            requests.Response: If a successful response is received, it is returned.

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """
        
        def attempt_host(
            host: ConnURL, 
            node_path: List[Tuple[str, Exception]]
        ) -> Optional['Response']:
            try:
                started_at_wall = time.time()
                started_at_perf = time.perf_counter()
                result = self.fetch_response_with_host(
                    host, method, uri, json = json, headers = headers, stream = stream
                )
                request_time_perf = time.perf_counter() - started_at_perf
                ended_at_wall = time.time()

                if (
                    query_info is not None
                    and self.log_config.slow_query is not None
                    and self.log_config.slow_query.get("enabled", True)
                ):
                    config = cast(
                        logging.SlowQueryLogMessageConfig,
                        self.log_config.slow_query,
                    )
                    if request_time_perf >= config.get("threshold_seconds", 5):
                        config["method"](
                            logging.QueryInfo(
                                operations=(
                                    query_info.operations() if callable(query_info.operations) else query_info.operations
                                ),
                                params=(
                                    query_info.params() if callable(query_info.params) else query_info.params
                                ),
                                request_type=(
                                    query_info.request_type() if callable(query_info.request_type) else query_info.request_type
                                ),
                                consistency=query_info.consistency,
                                freshness=query_info.freshness,
                            ),
                            duration_seconds=request_time_perf,
                            host = host.bare,
                            response_size_bytes = int(result.headers.get("Content-Length", "0")),
                            started_at = started_at_wall,
                            ended_at = ended_at_wall,
                        )

                return result
            
            except ConnectError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Failed to connect to node {e.host} - {str_error}"
                logging.log(self.log_config.connect_timeout, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except MaxRedirectsError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_redirect_path = str(e.redirect_path)
                    if max_length is not None and len(str_redirect_path) > max_length:
                        str_redirect_path = f"{str_redirect_path[:max_length]}..."
                    return f"Max redirects reached for node {e.host} (redirect path: {str_redirect_path})"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except UnexpectedResponse as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Unexpected response from node {e.host}: {str_error}"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                raise
        return self.try_hosts(attempt_host, initial_host=initial_host)

    @t.overload
    async def afetch_response(
        self,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: t.Literal[False] = False,
        initial_host: Optional[ConnURL] = None,
        query_info: Optional[logging.QueryInfoLazy] = None,
    ) -> 'Response':
        """Fetches a response from the server by requesting it from a random node. If
        a connection error occurs, this method will retry the request on a different
        node until it succeeds or the maximum number of attempts is reached.

        Args:
            method (str): The HTTP method to use for the request.
            uri (str): The URI of the request.
            json (dict): The json data to send with the request.
            headers (dict): The headers to send with the request.
            stream (bool): If True, the response will be streamed.
            initial_host (Optional[ConnURL]): The host to try first. If None,
                a random host will be chosen.
            query_info (Optional[logging.QueryInfo]): The query info for slow
                query logging, or None to disable slow query logging regardless of
                the log configuration.

        Returns:
            requests.Response: If a successful response is received, it is returned.

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """
        ...

    @t.overload
    async def afetch_response(
        self,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: t.Literal[True] = True,
        initial_host: Optional[ConnURL] = None,
        query_info: Optional[logging.QueryInfoLazy] = None,
    ) -> 'AsyncResponse':
        """Fetches a response from the server by requesting it from a random node. If
        a connection error occurs, this method will retry the request on a different
        node until it succeeds or the maximum number of attempts is reached.

        Args:
            method (str): The HTTP method to use for the request.
            uri (str): The URI of the request.
            json (dict): The json data to send with the request.
            headers (dict): The headers to send with the request.
            stream (bool): If True, the response will be streamed.
            initial_host (Optional[ConnURL]): The host to try first. If None,
                a random host will be chosen.
            query_info (Optional[logging.QueryInfo]): The query info for slow
                query logging, or None to disable slow query logging regardless of
                the log configuration.

        Returns:
            requests.Response: If a successful response is received, it is returned.

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """
        ...

    async def afetch_response(
        self,
        method: Literal["GET", "POST"],
        uri: str,
        json: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        stream: bool = False,
        initial_host: Optional[ConnURL] = None,
        query_info: Optional[logging.QueryInfoLazy] = None,
    ) -> 'Response':
        """Fetches a response from the server by requesting it from a random node. If
        a connection error occurs, this method will retry the request on a different
        node until it succeeds or the maximum number of attempts is reached.

        Args:
            method (str): The HTTP method to use for the request.
            uri (str): The URI of the request.
            json (dict): The json data to send with the request.
            headers (dict): The headers to send with the request.
            stream (bool): If True, the response will be streamed.
            initial_host (Optional[ConnURL]): The host to try first. If None,
                a random host will be chosen.
            query_info (Optional[logging.QueryInfo]): The query info for slow
                query logging, or None to disable slow query logging regardless of
                the log configuration.

        Returns:
            requests.Response: If a successful response is received, it is returned.

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """
        
        async def attempt_host(
            host: ConnURL, 
            node_path: List[Tuple[str, Exception]]
        ) -> Optional['Response']:
            try:
                started_at_wall = time.time()
                started_at_perf = time.perf_counter()
                result = await self.afetch_response_with_host(
                    host, method, uri, json = json, headers = headers, stream = stream
                )
                request_time_perf = time.perf_counter() - started_at_perf
                ended_at_wall = time.time()

                if (
                    query_info is not None
                    and self.log_config.slow_query is not None
                    and self.log_config.slow_query.get("enabled", True)
                ):
                    config = cast(
                        logging.SlowQueryLogMessageConfig,
                        self.log_config.slow_query,
                    )
                    if request_time_perf >= config.get("threshold_seconds", 5):
                        config["method"](
                            logging.QueryInfo(
                                operations=(
                                    query_info.operations() if callable(query_info.operations) else query_info.operations
                                ),
                                params=(
                                    query_info.params() if callable(query_info.params) else query_info.params
                                ),
                                request_type=(
                                    query_info.request_type() if callable(query_info.request_type) else query_info.request_type
                                ),
                                consistency=query_info.consistency,
                                freshness=query_info.freshness,
                            ),
                            duration_seconds=request_time_perf,
                            host = host.bare,
                            response_size_bytes = int(result.headers.get("Content-Length", "0")),
                            started_at = started_at_wall,
                            ended_at = ended_at_wall,
                        )

                return result
            
            except ConnectError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Failed to connect to node {e.host} - {str_error}"
                logging.log(self.log_config.connect_timeout, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except MaxRedirectsError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_redirect_path = str(e.redirect_path)
                    if max_length is not None and len(str_redirect_path) > max_length:
                        str_redirect_path = f"{str_redirect_path[:max_length]}..."
                    return f"Max redirects reached for node {e.host} (redirect path: {str_redirect_path})"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except UnexpectedResponse as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Unexpected response from node {e.host}: {str_error}"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                raise
        
        return await self.atry_hosts(attempt_host, initial_host=initial_host)

    def try_hosts(
        self,
        attempt_host: Callable[
            [ConnURL, List[Tuple[str, Exception]]], Optional[T]
        ],
        /,
        *,
        initial_host: Optional[ConnURL] = None,
    ) -> T:
        """Given a function like

        ```py
        def attempt_host(host: ConnURL, node_path: List[Tuple[str, Exception]]) -> Optional[str]:
            try:
                # do something with the host
                return 'result'
            except Exception as e: # only catch exceptions that should continue to the next host
                node_path.append((host, e))
                return None
        ```

        This will call `attempt_host` on each host in the connection in a random
        order until a non-None value is returned or the maximum number of attempts
        per host is reached.

        Args:
            attempt_host: The function to call on each host. It should return None
                if the host failed but we should continue to the next host, or a
                non-None value if the host succeeded. It should raise an exception if
                the host failed and we should not continue to the next host.
            initial_host: The host to try first. If None, a random host will be chosen.
                This does not need to be a host in self.hosts, but if it is not, it
                will not be included in the count for the maximum number of attempts
        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            Exception: any errors raised by `attempt_host`
        """
        node_path: List[Tuple[str, Exception]] = []
        first_node_idx: Optional[int] = None
        if initial_host is not None:
            try:
                first_node_idx = self.hosts.index(initial_host)
            except ValueError:
                if resp := attempt_host(initial_host, node_path):
                    return resp

        if first_node_idx is None: first_node_idx = random.randrange(0, len(self.hosts))
        if resp := attempt_host(self.hosts[first_node_idx], node_path): return resp
        remaining_nodes = [
            h for idx, h in enumerate(self.hosts) if idx != first_node_idx
        ]
        random.shuffle(remaining_nodes)
        node_ordering = [self.hosts[first_node_idx]] + remaining_nodes
        index, attempt = 0, 1

        while index < len(self.hosts) or attempt < self.max_attempts_per_host:
            if index + 1 < len(self.hosts):
                index += 1
            else:
                index = 0
                attempt += 1

            if resp := attempt_host(node_ordering[index], node_path): return resp
        raise MaxAttemptsError(node_path)
    

    async def atry_hosts(
        self,
        attempt_host: Callable[
            [ConnURL, List[Tuple[str, Exception]]], Awaitable[Optional[T]]
        ],
        /,
        *,
        initial_host: Optional[ConnURL] = None,
    ) -> T:
        """Given a function like

        ```py
        def attempt_host(host: ConnURL, node_path: List[Tuple[str, Exception]]) -> Optional[str]:
            try:
                # do something with the host
                return 'result'
            except Exception as e: # only catch exceptions that should continue to the next host
                node_path.append((host, e))
                return None
        ```

        This will call `attempt_host` on each host in the connection in a random
        order until a non-None value is returned or the maximum number of attempts
        per host is reached.

        Args:
            attempt_host: The function to call on each host. It should return None
                if the host failed but we should continue to the next host, or a
                non-None value if the host succeeded. It should raise an exception if
                the host failed and we should not continue to the next host.
            initial_host: The host to try first. If None, a random host will be chosen.
                This does not need to be a host in self.hosts, but if it is not, it
                will not be included in the count for the maximum number of attempts
        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            Exception: any errors raised by `attempt_host`
        """
        node_path: List[Tuple[str, Exception]] = []
        first_node_idx: Optional[int] = None
        if initial_host is not None:
            try:
                first_node_idx = self.hosts.index(initial_host)
            except ValueError:
                if resp := await attempt_host(initial_host, node_path):
                    return resp

        if first_node_idx is None: first_node_idx = random.randrange(0, len(self.hosts))
        if resp := await attempt_host(self.hosts[first_node_idx], node_path): return resp
        remaining_nodes = [
            h for idx, h in enumerate(self.hosts) if idx != first_node_idx
        ]
        random.shuffle(remaining_nodes)
        node_ordering = [self.hosts[first_node_idx]] + remaining_nodes
        index, attempt = 0, 1

        while index < len(self.hosts) or attempt < self.max_attempts_per_host:
            if index + 1 < len(self.hosts):
                index += 1
            else:
                index = 0
                attempt += 1

            if resp := await attempt_host(node_ordering[index], node_path): return resp
        raise MaxAttemptsError(node_path)
    
    def discover_leader(self) -> ConnURL:
        """Discovers the current leader for the cluster

        Returns:
            A tuple of (leader_host, leader_port)

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """

        def attempt_host(
                host: ConnURL, 
                node_path: List[Tuple[str, Exception]]
            ) -> Optional[ConnURL]:
            try:
                return self.discover_leader_with_host(host)
            
            except ConnectError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Failed to connect to node {e.host} - {str_error}"
                logging.log(self.log_config.connect_timeout, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except UnexpectedResponse as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Unexpected response from node {e.host}: {str_error}"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                raise

        return self.try_hosts(attempt_host)
    

    async def adiscover_leader(self) -> ConnURL:
        """Discovers the current leader for the cluster

        Returns:
            A tuple of (leader_host, leader_port)

        Raises:
            MaxAttemptsError: If the maximum number of attempts is reached before
                we get a successful response.
            UnexpectedResponse: If one of the rqlite nodes returns a response
                we didn't expect
        """

        async def attempt_host(
                host: ConnURL, 
                node_path: List[Tuple[str, Exception]]
            ) -> Optional[ConnURL]:
            try:
                return await self.adiscover_leader_with_host(host)
            
            except ConnectError as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Failed to connect to node {e.host} - {str_error}"
                logging.log(self.log_config.connect_timeout, msg_supplier, exc_info=True)
                node_path.append((e.host, e))
            
            except UnexpectedResponse as e:
                def msg_supplier(max_length: Optional[int]) -> str:
                    str_error = str(e)
                    if max_length is not None and len(str_error) > max_length: str_error = f"{str_error[:max_length]}..."
                    return f"Unexpected response from node {e.host}: {str_error}"

                logging.log(self.log_config.non_ok_response, msg_supplier, exc_info=True)
                raise

        return self.try_hosts(attempt_host)

    def discover_leader_with_host(self, host: ConnURL) -> ConnURL:
        """Uses the given node in the cluster to discover the current leader
        for the cluster.

        Returns:
            A tuple of (leader_host, leader_port)

        Raises:
            ConnectTimeout: If the connection times out.
            UnexpectedResponse: If the server returns a response we didn't expect.
        """
        response = None
        try:
            response = aioreq.request(
                "POST",
                host.with_uri("/db/query?level=weak&redirect"),
                json = [["SELECT 1"]],
                headers = {"Content-Type": "application/json; charset=UTF-8"},
                timeout = self.timeout,
                allow_redirects = False,
            )

            if response.is_redirect:
                redirected_to = response.headers["Location"]
                parsed_url = urllib.parse.urlparse(redirected_to)
                return ConnURL(host = parse_host(
                    parsed_url.netloc,
                    # default_port = 443 if parsed_url.scheme == "https" else 80,
                ))

            if response.status_code < 200 or response.status_code > 299:
                raise UnexpectedResponse(
                    host.safe_url,
                    f"Unexpected response from {host.bare}: {response.status_code} {response.reason}",
                )

            return host
        except self._excs.ConnectTimeout as e:
            raise ConnectError(
                f"Connection to {host.bare} timed out", host.bare
            ) from e
        except self._excs.ConnectionError as e:
            raise ConnectError(
                f"Connection to {host.bare} was refused", f"{host.bare}"
            ) from e
        

    async def adiscover_leader_with_host(self, host: ConnURL) -> ConnURL:
        """Uses the given node in the cluster to discover the current leader
        for the cluster.

        Returns:
            A tuple of (leader_host, leader_port)

        Raises:
            ConnectTimeout: If the connection times out.
            UnexpectedResponse: If the server returns a response we didn't expect.
        """
        response = None
        try:
            response = await aioreq.arequest(
                "POST",
                host.with_uri("/db/query?level=weak&redirect"),
                json = [["SELECT 1"]],
                headers = {"Content-Type": "application/json; charset=UTF-8"},
                timeout = self.timeout,
                allow_redirects = False,
            )

            if response.is_redirect:
                redirected_to = response.headers["Location"]
                parsed_url = urllib.parse.urlparse(redirected_to)
                return ConnURL(host = parse_host(
                    parsed_url.netloc,
                    # default_port = 443 if parsed_url.scheme == "https" else 80,
                ))

            if response.status_code < 200 or response.status_code > 299:
                raise UnexpectedResponse(
                    host.safe_url,
                    f"Unexpected response from {host.bare}: {response.status_code} {response.reason}",
                )

            return host
        except self._excs.ConnectTimeout as e:
            raise ConnectError(
                f"Connection to {host.bare} timed out", host.bare
            ) from e
        except self._excs.ConnectionError as e:
            raise ConnectError(
                f"Connection to {host.bare} was refused", f"{host.bare}"
            ) from e


from __future__ import annotations

"""
Class Object Wrappers
"""
import copy
import inspect
import functools
import contextlib
import typing as t
try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

SelfT = t.TypeVar("SelfT")
ParamT = ParamSpec("ParamT")
ReturnType = t.TypeVar("ReturnType", covariant=True)
# https://stackoverflow.com/questions/77540153/type-hinting-decorator-which-injects-the-value-but-also-supports-passing-the-va?noredirect=1&lq=1

_is_async_default: t.Optional[bool] = None
_curr_async_default: t.Optional[bool] = None

def set_default_is_async(
    default: t.Optional[bool] = None,
):
    """
    Sets the default value for the `is_async` kwarg
    """
    global _is_async_default
    _is_async_default = default

@contextlib.contextmanager
def syncoro_ctx(
    is_async: t.Optional[bool] = None,
):
    """
    Context Manager that allows setting the `is_async` kwarg
    """
    global _curr_async_default
    old_is_async = _curr_async_default
    if is_async is not None: _curr_async_default = is_async
    yield
    _curr_async_default = old_is_async

def get_default_is_async(
    _is_async: t.Optional[bool] = None,
    _default: t.Optional[bool] = None,
) -> bool:
    """
    Returns the default value for the `is_async` kwarg
    """
    if _curr_async_default is not None: return _curr_async_default
    if _is_async is not None: return _is_async
    return _default if _default is not None else _is_async_default

def syncoro_cls(
    fn: t.Callable[ParamT, ReturnType],
    async_fn: t.Callable[ParamT, t.Awaitable[ReturnType]],
    _default: t.Optional[bool] = None,
):
    """
    Creates a classmethod Wrapper that allows for both sync and async calls
    based on the presence of the `is_async` kwarg
    """
    
    fn = fn.__func__ if isinstance(fn, classmethod) else fn
    async_fn = async_fn.__func__ if isinstance(async_fn, classmethod) else async_fn
    def inner(func: t.Callable[ParamT, ReturnType]) -> t.Callable[ParamT, ReturnType | t.Awaitable[ReturnType]]:
        """
        Inner Wrapper
        """
        # _default = _is_async_default if _default is None else _default

        @functools.wraps(func)
        def wrapper(cls, *args: ParamT.args, is_async: t.Optional[bool] = None, **kwargs: ParamT.kwargs) -> ReturnType | t.Awaitable[ReturnType]:
            is_async = get_default_is_async(is_async, _default)
            # if is_async is None: is_async = _default if _default is not None else _is_async_default
            return async_fn(cls, *args, **kwargs) if is_async else fn(cls, *args, **kwargs)
        return classmethod(wrapper)
    return inner


class Syncoro(t.Protocol):
    def __call__(self, *args: ParamT.args, is_async: t.Optional[bool] = None, **kwargs: ParamT.kwargs) -> ReturnType | t.Awaitable[ReturnType]:
        ...
        
def syncoro(
    fn: t.Callable[ParamT, ReturnType],
    async_fn: t.Callable[ParamT, t.Awaitable[ReturnType]],
    _default: t.Optional[bool] = None,
    _is_classmethod: t.Optional[bool] = None,
# ) -> t.Callable[[t.Callable[..., ReturnType]], t.Callable[ParamT, ReturnType]]:
):
    """
    Creates a Wrapper that allows for both sync and async calls
    based on the presence of the `is_async` kwarg
    """
    if _is_classmethod: return syncoro_cls(fn, async_fn, _default)
    
    def inner(func: t.Callable[ParamT, ReturnType]) -> t.Callable[ParamT, ReturnType | t.Awaitable[ReturnType]]:
    # def inner(func: t.Callable[ParamT, ReturnType]) -> t.Callable[t.Concatenate[ParamT, (bool | None, "is_async")], ReturnType | t.Awaitable[ReturnType]]:
    # def inner(func: t.Callable[ParamT, ReturnType]):
        """
        Inner Wrapper
        """

        @functools.wraps(func)
        def wrapper(*args: ParamT.args, is_async: t.Optional[bool] = None, **kwargs: ParamT.kwargs) -> ReturnType | t.Awaitable[ReturnType]:
            is_async = get_default_is_async(is_async, _default)
            # if is_async is None: is_async = _default if _default is not None else _is_async_default
            return async_fn(*args, **kwargs) if is_async else fn(*args, **kwargs)
        
        # is_async_param = inspect.Parameter('is_async', inspect.Parameter.KEYWORD_ONLY, default = _default, annotation = t.Optional[bool])
        # sig = inspect.signature(func)
        # params = list(sig.parameters.values())
        # params.append(is_async_param)
        
        # # params.insert(len(params) - 1, is_async_param)
        # # wrapper: Syncoro
        # wrapper.__signature__ = sig.replace(parameters = params)
        # wrapper.__annotations__ = copy.deepcopy(func.__annotations__)
        # rt = wrapper.__annotations__.pop('return')
        # wrapper.__annotations__['is_async'] = bool
        # wrapper.__annotations__['return'] = rt
        return wrapper
    return inner



from __future__ import annotations

"""
A super simple wrapper for `asyncio.run` or `anyio.run`
that also handles adding in hooks
"""
import inspect
import asyncio
import functools
import typing as t

try:
    import anyio
    _has_anyio = True
except ImportError:
    _has_anyio = False

HookT = t.Callable[[], t.Awaitable[None]]
_init_hooks: t.Set[HookT] = set()
_finalize_hooks: t.Set[HookT] = set()
_added_tasks: t.Set[asyncio.Task] = set()

def add_init_hook(func: HookT | t.Sequence[HookT]):
    """
    Adds an init hook(s)
    """
    if isinstance(func, list): _init_hooks.update(func)
    else: _init_hooks.add(func)
        
def add_finalize_hook(func: HookT | t.Sequence[HookT]):
    """
    Adds a finalize hook(s)
    """
    if isinstance(func, list): _finalize_hooks.update(func)
    else: _finalize_hooks.add(func)

def create_task(coro: t.Callable[..., t.Awaitable[t.Any]] | t.Awaitable[t.Any], *args, _task_list: t.Optional[t.Set[asyncio.Task]] = None, _task_name: t.Optional[str] = None, **kwargs) -> asyncio.Task:
    """
    Adds a task to the finalize hooks
    """
    if not inspect.isawaitable(coro): coro = coro(*args, **kwargs)
    task = asyncio.create_task(coro, name = _task_name)
    _added_tasks.add(task)
    if _task_list is not None: 
        _task_list.add(task)
        task.add_done_callback(_task_list.discard)
    # Remove the task  from _added_tasks
    task.add_done_callback(_added_tasks.discard)
    return task

async def complete_added_tasks(task_list: t.Optional[t.Set[asyncio.Task]] = None):
    """
    Completes all current tasks
    """
    if task_list is not None: await asyncio.gather(*task_list)
    elif _added_tasks: await asyncio.gather(*_added_tasks)

async def runtime(
    func: t.Callable[..., t.Awaitable[t.Any]] | t.Awaitable[t.Any],
    *args,
    **kwargs,
):
    """
    Runs the function with the runtime
    """
    for hook in _init_hooks:
        await hook(*args, **kwargs)
    try:
        return await func if inspect.isawaitable(func) else await func(*args, **kwargs)
    finally:
        if _added_tasks: await asyncio.gather(*_added_tasks)
        for hook in _finalize_hooks:
            await hook(*args, **kwargs)
        

def run(
    func: t.Callable[..., t.Awaitable[t.Any]] | t.Awaitable[t.Any],
    *args,
    _runtime: t.Optional[t.Literal['asyncio', 'anyio']] = None,
    **kwargs,
):
    """
    Runs the function with the runtime
    """
    if _runtime is None:
        _runtime = 'anyio' if _has_anyio else 'asyncio'
    if _runtime == 'asyncio':
        return asyncio.run(runtime(func, *args, **kwargs))
    if _runtime == 'anyio':
        func_partial = functools.partial(runtime, func, *args, **kwargs)
        return anyio.run(func_partial)
    raise ValueError(f'Invalid Runtime: {_runtime}')





from __future__ import annotations

import asyncio
import typing as t
import dataclasses
from temporalio.converter import DataConverter
from temporalio.client import Client, KeepAliveConfig, WorkflowExecutionStatus
from .utils import logger
from . import patches

if t.TYPE_CHECKING:
    from temporalio.common import QueryRejectCondition
    from temporalio.runtime import Runtime
    from temporalio.client import (
        TLSConfig,
        Interceptor,
        RetryConfig,
        HttpConnectProxyConfig,
        MethodAsyncSingleParam,
        ParamType,
        SelfType, ParamType, ReturnType,

    )
    from temporalio.worker import Worker
    from lzl.ext.temporal.configs import TemporalSettings



    
class TemporalClient(Client):

    
    # @staticmethod
    @classmethod
    async def connect(
        cls: t.Type['TemporalClient'],
        target_host: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        api_key: t.Optional[str] = None,
        data_converter: t.Optional['DataConverter'] = None,
        interceptors: t.Sequence['Interceptor'] = [],
        default_workflow_query_reject_condition: t.Optional[
            'QueryRejectCondition'
        ] = None,
        tls: t.Union[bool, 'TLSConfig'] = False,
        retry_config: t.Optional['RetryConfig'] = None,
        keep_alive_config: t.Optional['KeepAliveConfig'] = KeepAliveConfig.default,
        rpc_metadata: t.Mapping[str, str] = {},
        identity: t.Optional[str] = None,
        lazy: bool = False,
        runtime: t.Optional['Runtime'] = None,
        http_connect_proxy_config: t.Optional['HttpConnectProxyConfig'] = None,
        default_task_queue: t.Optional[str] = None,
        config: t.Optional['TemporalSettings'] = None,
        **kwargs,
    ) -> TemporalClient:  # sourcery skip: default-mutable-arg
        """Connect to a Temporal server.

        Args:
            target_host: ``host:port`` for the Temporal server. For local
                development, this is often "localhost:7233".
            namespace: Namespace to use for client calls.
            api_key: API key for Temporal. This becomes the "Authorization"
                HTTP header with "Bearer " prepended. This is only set if RPC
                metadata doesn't already have an "authorization" key.
            data_converter: Data converter to use for all data conversions
                to/from payloads.
            interceptors: Set of interceptors that are chained together to allow
                intercepting of client calls. The earlier interceptors wrap the
                later ones.

                Any interceptors that also implement
                :py:class:`temporalio.worker.Interceptor` will be used as worker
                interceptors too so they should not be given when creating a
                worker.
            default_workflow_query_reject_condition: The default rejection
                condition for workflow queries if not set during query. See
                :py:meth:`WorkflowHandle.query` for details on the rejection
                condition.
            tls: If false, the default, do not use TLS. If true, use system
                default TLS configuration. If TLS configuration present, that
                TLS configuration will be used.
            retry_config: Retry configuration for direct service calls (when
                opted in) or all high-level calls made by this client (which all
                opt-in to retries by default). If unset, a default retry
                configuration is used.
            keep_alive_config: Keep-alive configuration for the client
                connection. Default is to check every 30s and kill the
                connection if a response doesn't come back in 15s. Can be set to
                ``None`` to disable.
            rpc_metadata: Headers to use for all calls to the server. Keys here
                can be overriden by per-call RPC metadata keys.
            identity: Identity for this client. If unset, a default is created
                based on the version of the SDK.
            lazy: If true, the client will not connect until the first call is
                attempted or a worker is created with it. Lazy clients cannot be
                used for workers.
            runtime: The runtime for this client, or the default if unset.
            http_connect_proxy_config: Configuration for HTTP CONNECT proxy.
            default_task_queue: The default task queue to use for this client.
            config: The configuration for this client.
        """
        if config is None:
            from lzl.ext.temporal.configs import get_temporal_settings
            config = get_temporal_settings()
        
        if not target_host: target_host = config.host
        # print('target_host: ', target_host)
        if not namespace and config.namespace: namespace = config.namespace
        if not api_key and config.api_key: api_key = config.api_key
        if not default_task_queue and config.default_task_queue: default_task_queue = config.default_task_queue
        if not tls and config.tls is not None: tls = config.tls
        if not identity and config.identity: identity = config.identity
        if not lazy and config.lazy is not None: lazy = config.lazy
        if data_converter is None: data_converter = config.data_converter

        from temporalio.service import ServiceClient, ConnectConfig
        connect_config = ConnectConfig(
            target_host=target_host,
            api_key=api_key,
            tls=tls,
            retry_config=retry_config,
            keep_alive_config=keep_alive_config,
            rpc_metadata=rpc_metadata,
            identity=identity or "",
            lazy=lazy,
            runtime=runtime,
            http_connect_proxy_config=http_connect_proxy_config,
        )
        new = cls(
            await ServiceClient.connect(connect_config),
            namespace=namespace,
            data_converter=data_converter,
            interceptors=interceptors,
            default_workflow_query_reject_condition=default_workflow_query_reject_condition,
        )
        new._postinit_config_(config = config, **kwargs)
        return new

    def _postinit_config_(self, config: t.Optional['TemporalSettings'] = None, **kwargs) -> None:
        """
        Some post-init config
        """
        if config is None:
            from lzl.ext.temporal.configs import get_temporal_settings
            config = get_temporal_settings()
        self.tmprl_config = config
        self.tmprl_registry = self.tmprl_config.registry
        self._extra: t.Dict[str, t.Any] = {}
        self.tmprl_registry.register_client(self)
        # self.tmprl_registry.clients[self.namespace] = self
        # self.tmprl_registry.client = self
    
    async def run_worker(
        self,
        worker: 'Worker',
        event: asyncio.Event,
        **kwargs,
    ):
        """
        Runs a Temporal Worker
        """
        from lzo.utils import Timer
        ts = Timer(format_short = 1)
        extra = f'|g|NS|e|: `{self.namespace}`' if self.namespace else ''
        if worker.task_queue: extra += f', |g|TQ|e|: `{worker.task_queue}`'
        extra = f' ({extra.strip()})' if extra else ''
        logger.info(f'Starting Temporal Worker{extra}', colored = True)
        await worker.run()
        try:
            await event.wait()
        finally:
            logger.info(f'Shutting Down Temporal Worker. (|g|TTL|e|: {ts.total_s})', prefix = worker.task_queue, colored = True)
        
    # def run_worker(
    #     self,
    #     worker: 'Worker',
    #     event: t.Optional[asyncio.Event] = None,
    #     **kwargs,
    # ):
    #     """
    #     Runs a Temporal Worker
    #     """
    #     if event is None: event = asyncio.Event()
    #     loop = asyncio.get_running_loop()
    #     try:

    #         loop.run_until_complete(self._run_worker(worker, event, **kwargs))
    #     except Exception as e:
    #         logger.error(f'Error Running Temporal Worker: {e}')
    #         raise e
    #     finally:
    #         event.set()
    #         loop.run_until_complete(loop.shutdown_asyncgens())
    if t.TYPE_CHECKING:
        def get_workflow_handle(
            self,
            workflow_id: str,
            *,
            run_id: t.Optional[str] = None,
            first_execution_run_id: t.Optional[str] = None,
            result_type: t.Optional[t.Type] = None,
        ) -> patches.WorkflowHandle[t.Any, t.Any]:
            """Get a workflow handle to an existing workflow by its ID.

            Args:
                workflow_id: Workflow ID to get a handle to.
                run_id: Run ID that will be used for all calls.
                first_execution_run_id: First execution run ID used for cancellation
                    and termination.
                result_type: The result type to deserialize into if known.

            Returns:
                The workflow handle.
            """
            ...


        

    if not t.TYPE_CHECKING:
        # async def start_workflow(
        #     self,
        #     workflow: t.Callable[
        #         t.Concatenate['SelfType', 'MultiParamSpec'], t.Awaitable['ReturnType']
        #     ],
        #     arg: t.Any, 
        #     *,
        #     args: t.Sequence[t.Any] = [], 
        #     id: str,
        #     task_queue: t.Optional[str] = None,
        #     **kwargs,
        # ):
        #     if task_queue is None and self._tconf.default_task_queue: task_queue = self._tconf.default_task_queue
        #     return await super().start_workflow(workflow, arg, args = args, task_queue = task_queue, id = id, **kwargs)

        async def execute_workflow(
            self,
            workflow: 'MethodAsyncSingleParam[SelfType, ParamType, ReturnType]',
            arg: t.Any, 
            *,
            args: t.Sequence[t.Any] = [], 
            id: str,
            task_queue: t.Optional[str] = None,
            **kwargs,
        ):
            if task_queue is None and self.tmprl_config.default_task_queue: task_queue = self.tmprl_config.default_task_queue
            return await super().execute_workflow(workflow, arg, args = args,  task_queue = task_queue, id = id, **kwargs)
        






from __future__ import annotations

import typing as t
from lzo.types import BaseModel, Field, eproperty

class ResultItem(BaseModel):
    """
    The Single Search Result Item
    """
    url: str = Field(..., description="The URL of the search result")
    title: str = Field(..., description="The title of the search result")
    query: str = Field(..., description="The query used to obtain this search result")
    content: t.Optional[str] = Field(None, description="The content snippet of the search result")




import os
from lzl.logging import get_logger, change_logger_level, null_logger
logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()
logger = get_logger(logger_level)

logger.set_module_name('lzl.api.aioreq', 'aioreq', is_relative = True)



from __future__ import annotations

"""
Temporal Workflow Mixins
"""


import uuid
import typing as t

if t.TYPE_CHECKING:
    from .base import BaseTemporalMixin


def default_id_gen_function(obj: 'BaseTemporalMixin', *args, prefix: t.Optional[str] = None, **kwargs) -> str:
    """
    Default ID Generator Function
    """
    # logger.info(f'ID Generator Options: {obj}', prefix = 'Temporal')
    base = prefix or obj.id_gen.prefix or ''
    if obj.id_gen.func: id_gen = obj.id_gen.func(*args, **kwargs)
    else: id_gen = str(uuid.uuid4().int)
    # id_gen = str(uuid.uuid4().int)
    if obj.id_gen.id_length: id_gen = id_gen[:obj.id_gen.id_length]
    base += f'{obj.id_gen.joiner}{id_gen}'
    if obj.id_gen.suffix: base += f'{obj.id_gen.joiner}{obj.id_gen.suffix}'
    base = base.lstrip(obj.id_gen.joiner).rstrip(obj.id_gen.joiner)
    base = base.replace(" ", obj.id_gen.joiner).replace(f'{obj.id_gen.joiner}{obj.id_gen.joiner}', obj.id_gen.joiner)
    if obj.id_gen.lower: base = base.lower()
    if obj.id_gen.max_length and len(base) > obj.id_gen.max_length: base = base[:obj.id_gen.max_length]
    return base


from __future__ import annotations

"""
Client Utilities
"""

import inspect
import backoff
import functools
import random
from lzl import load
from typing import Optional, Type, TypeVar, Union, Set, List, Any, Dict, Literal, Callable, TYPE_CHECKING

if load.TYPE_CHECKING:
    import aiohttpx
    from aiohttpx import HTTPStatusError
else:
    aiohttpx = load.LazyLoad("aiohttpx", install_missing=True)

def create_cachify_build_name_func(
    base_name: str,
    special_names: Optional[List[str]] = None,
    function_names: Optional[List[str]] = None,
    include_classname: Optional[bool] = True,
    include_http_methods: Optional[bool] = False,
) -> Callable[..., str]:
    """
    Creates a function that returns the build name
    """

    def build_name_func(func: Callable, *args, **kwargs) -> str:
        """
        Build the name function
        """
        nonlocal base_name
        func = inspect.unwrap(func)
        func_name = func.__qualname__ if include_classname else func.__name__

        if special_names:
            for name in special_names:
                if name in func_name.lower():
                    func_name = func_name.replace(f'a{name}_', '').replace(f'{name}_', '').replace('__', '_')
                    return f"{base_name}.{name}_{func_name}"
        
        if function_names:
            for name in function_names:
                if name in func_name.lower(): return f"{base_name}.{func_name}"
        
        if include_http_methods:
            for method in {'get', 'post', 'put', 'delete'}:
                if method in func_name.lower(): return f"{base_name}.{method}"
        return f"{base_name}.{func_name}"
    return build_name_func


def fatal_http_exception(exc: Union[Exception, 'HTTPStatusError']) -> bool:
    """
    Checks if the exception is fatal
    """
    if isinstance(exc, aiohttpx.HTTPStatusError):
        # retry on server errors and client errors
        # with 429 status code (rate limited),
        # don't retry on other client errors
        return (400 <= exc.response.status_code < 500) and exc.response.status_code != 429
    else:
        # retry on all other errors (eg. network)
        return False


http_retry_wrapper = functools.partial(
    backoff.on_exception,
    backoff.expo, 
    exception = Exception, 
    giveup = fatal_http_exception,
    factor = 5,
)

DefaultUserAgents = [
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36',
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 Edg/116.0.1938.81",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 Edg/119.0.0.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.3",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.3",
]

def get_user_agent(*args, **kwargs) -> str:
    """
    Returns a user agent
    """
    try:
        from fake_useragent import UserAgent
        ua = UserAgent(*args, **kwargs)
        return ua.random
    except ImportError:
        return random.choice(DefaultUserAgents)

try:
    from browserforge.headers import HeaderGenerator
    _bheaders = True
    _bheader_gen: Optional[HeaderGenerator] = None
except ImportError:
    _bheaders = False
    _bheader_gen = None


def get_default_headers(browser: Optional[str] = 'chrome', **kwargs) -> Dict[str, str]:
    """
    Returns the default headers
    """
    global _bheader_gen
    if _bheaders:
        if _bheader_gen is None: _bheader_gen = HeaderGenerator()
        return _bheader_gen.generate(browser = browser, **kwargs)
    return {}




from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from pydantic import BaseModel, AliasChoices


ModelT = t.TypeVar("ModelT", bound="BaseModel")

_extracted_pydantic_model_kws: t.Dict[str, t.List[str]] = {}

def extract_pydantic_model_kws(
    model: ModelT,
    kwargs: t.Dict[str, t.Any],
) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
    """
    Extracts the kwargs from the resource and returns the kwargs and the model kwargs
    """
    global _extracted_pydantic_model_kws
    base_model_name = f"{model.__module__}.{model.__name__}"
    if base_model_name not in _extracted_pydantic_model_kws:
        from .base import get_pydantic_field_names
        resource_kws = get_pydantic_field_names(model)
        _extracted_pydantic_model_kws[base_model_name] = resource_kws
    model_kwargs = {
        key: kwargs.pop(key) for key in kwargs if key in _extracted_pydantic_model_kws[base_model_name]
    }
    return kwargs, model_kwargs


def create_alias_choices(
    key: str,
    *values: str,
    key_pos: t.Literal['prefix', 'suffix'] = 'suffix',
    case: t.Optional[t.Literal['upper', 'lower']] = 'upper',
) -> 'AliasChoices':
    """
    Helper Function to create alias choices
    
    Args:
        key (str): The key to use for the alias choices
        *values (str): The values to use for the alias choices
        key_pos (Literal['prefix', 'suffix']): The position for the key to use for the alias choices

    Example:
    
        >>> create_alias_choices('database', 'url', 'port', key_pos = 'prefix')
        
        AliasChoices('DATABASE_URL', 'DATABASE_PORT')

        >>> create_alias_choices('url', 'redis', 'pg', key_pos = 'suffix')

        AliasChoices('REDIS_URL', 'PG_URL')

    """
    alias_values: t.List[str] = []
    if key_pos == 'suffix':
        alias_values.extend(f'{value}_{key}' for value in values)
    elif key_pos == 'prefix':
        alias_values.extend(f'{key}_{value}' for value in values)
    else: raise ValueError(f'Invalid Key Position: {key_pos}')
    if case == 'upper': alias_values = [v.upper() for v in alias_values]
    elif case == 'lower': alias_values = [v.lower() for v in alias_values]
    from pydantic import AliasChoices
    return AliasChoices(*alias_values)


from __future__ import annotations

from typing import Any, Dict, Type, Union


def dict_diff(dict_a: Dict[str, Any], dict_b: Dict[str, Any], show_value_diff: bool = True):
    
    """
    Compare two dictionaries and return the difference between them
    """
    result = {
        'added': {k: dict_b[k] for k in set(dict_b) - set(dict_a)},
        'removed': {k: dict_a[k] for k in set(dict_a) - set(dict_b)},
    }
    if show_value_diff:
        common_keys =  set(dict_a) & set(dict_b)
        result['value_diffs'] = {
            k:(dict_a[k], dict_b[k])
            for k in common_keys
            if dict_a[k] != dict_b[k]
        }
    return result


from __future__ import annotations

"""
Oso Exceptions
"""

from lzl.ext.fast.types.errors import FastAPIException

class OsoException(FastAPIException):
    """
    Oso Exception
    """
    base: str = "Oso Error"
    kind: str = "oso_error"
    status: int = 400
    problem: str = "Oso Error"
    solution: str = "Please check the error message for more information"
    module: str = "oso"

class ForbiddenError(OsoException):
    """
    Forbidden Error
    """
    status: int = 403
    kind = "fobidden_action"
    problem: str = "The requested action was not allowed for {actor} for the given resource [{resource}]"
    solution: str = "Please ensure that the proper permissions are set for the resource for the requested action: {action}"

class NotFoundError(OsoException):
    """
    Not Found Error
    """
    status: int = 404
    kind = "not_found"
    problem: str = "The current {actor} does not have permission for the given resource [{resource}]"
    solution: str = "Please ensure that the proper permissions are set for the resource for the requested action: {action}"



from __future__ import annotations

"""
Extra Helpers
"""
import functools
import typing as t
import urllib.request
import shutil
from lzl import load
from concurrent.futures import ThreadPoolExecutor, as_completed

if t.TYPE_CHECKING:
    import browserforge.download
    from lzl.api import aiohttpx
    from pydantic import BaseModel
else:
    browserforge = load.LazyLoad("browserforge", install_missing=True)
    aiohttpx = load.lazy_load('lzl.api.aiohttpx', install_missing = False)


def download_file(url: str, path: str) -> None:
    """
    Download a file from the specified URL and save it to the given path.
    """
    with urllib.request.urlopen(url) as resp:  # nosec
        if resp.status != 200:
            raise ValueError(f"Download failed with status code: {resp.status}")
        with open(path, "wb") as f:
            shutil.copyfileobj(resp, f)

@functools.lru_cache()
def browserforge_download(
    headers: t.Optional[bool] = True,
    fingerprints: t.Optional[bool] = True,
):
    """
    Returns the browser forge downloader
    """
    browserforge.__load__()
    from lzl.logging import logger
    try:
        from browserforge.download import is_downloaded
        if is_downloaded(headers = headers, fingerprints = fingerprints):
            return True
    except ImportError:
        from browserforge.download import IsDownloaded
        try:
            if IsDownloaded(): return True
        except Exception as e:
            logger.error(f'Error checking if downloaded: {e}')
            try:
                from browserforge.download import get_all_paths
            except ImportError:
                from browserforge.download import _get_all_paths as get_all_paths
            def _IsDownloaded() -> bool:
                """
                Check if the required data files are already downloaded.
                Returns True if all the requested data files are present, False otherwise.
                """
                for path in get_all_paths():
                    if not path.exists():
                        return False
                return True
            
            if _IsDownloaded(): return True


    from browserforge.download import (
        ROOT_DIR,
        DATA_DIRS,
        DATA_FILES,
        REMOTE_PATHS,
    )
    futures = {}
    options = {'headers': headers, 'fingerprints': fingerprints}
    options = {k: v for k, v in options.items() if v}
    with ThreadPoolExecutor(10) as executor:
        for data_type in options:
            for local_name, remote_name in DATA_FILES[data_type].items():
                url = f"{REMOTE_PATHS[data_type]}/{remote_name}"
                path = str(DATA_DIRS[data_type] / local_name)
                future = executor.submit(download_file, url, path)
                futures[future] = local_name
        for f in as_completed(futures):
            try:
                future.result()
                logger.info(f"{futures[f]:<30}OK!")
            except Exception as e:
                logger.warning(f"Error downloading {local_name}: {e}")



_json_schemas: t.Dict[str, str] = {}

def generate_json_schema_from_model(model: t.Type['BaseModel']) -> str:
    """
    Generates a JSON schema from a Pydantic model.
    
    Args:
        model (t.Type[BaseModel]): The Pydantic model class to generate the schema from.
    
    Returns:
        str: The generated JSON schema as a string.
    """
    if model.__name__ in _json_schemas:
        return _json_schemas[model.__name__]
    from lzl.api.openai.utils.schemas import generate_json_schema
    schema = generate_json_schema(model)
    _json_schemas[model.__name__] = schema
    return schema


from __future__ import annotations

import json
import pathlib
from lzl.logging import logger
from lzl.proxied import ProxyObject
from lzo.utils import load_env_vars
from lzo.types import BaseSettings, eproperty, model_validator, field_validator
from lzo.registry.mixins import RegisteredSettings
from .providers import (
    ConfigMixin,
    AWSConfig,
    GCPConfig,
    MinioConfig,
    S3CompatConfig,
    CloudflareR2Config,
)
from typing import Optional, Dict, Any, List, Union, Tuple, TYPE_CHECKING

ProviderConfig = Union[AWSConfig, GCPConfig, MinioConfig, S3CompatConfig, CloudflareR2Config]

class FileIOConfig(BaseSettings, RegisteredSettings.configure_registered(module = 'file')):
    """
    FileIOConfig
    """

    read_chunk_size: Optional[int] = 1024 * 64 # 64KB
    url_chunk_size: Optional[int] = 1024 * 128 # 128KB
    max_pool_connections: Optional[int] = 100
    
    num_workers: Optional[int] = 12
    checksum_cache_ttl: Optional[int] = 60 * 60 * 24 * 1 # 1 days
    enable_progress_bar: Optional[bool] = False
    tfio_enabled: Optional[bool] = False

    class Config:
        env_prefix = "FILEIO_"
        case_sensitive = False

    @eproperty
    def aws(self) -> AWSConfig:
        """
        Returns the AWS Config
        """
        return AWSConfig()
    
    @eproperty
    def gcp(self) -> GCPConfig:
        """
        Returns the GCP Config
        """
        return GCPConfig()
    
    @eproperty
    def minio(self) -> MinioConfig:
        """
        Returns the Minio Config
        """
        return MinioConfig()
    
    @eproperty
    def s3c(self) -> S3CompatConfig:
        """
        Returns the S3Compat Config
        """
        return S3CompatConfig()
    
    @eproperty
    def r2(self) -> CloudflareR2Config:
        """
        Returns the Cloudflare R2 Config
        """
        return CloudflareR2Config()
    

    def create_adc(
        self, 
        data: Union[str, Dict[str, Any]], 
        path: str = None
    ):
        """
        Create a new ADC based on the passed data and writes it to 
        path or GOOGLE_APPLICATION_CREDENTIALS
        """
        if isinstance(data, str): data = json.loads(data)
        path: pathlib.Path = pathlib.Path(path) if path else self.gcp.google_application_credentials
        path.write_text(json.dumps(data, indent = 2, ensure_ascii=False))
        self.gcp.google_application_credentials = path


    def get_boto_values(self):
        """
        Returns the boto config values
        """
        t = "[Credentials]\n"
        if self.aws.aws_access_key_id:
            t += f"aws_access_key_id = {self.aws.aws_access_key_id}\n"
        if self.aws.aws_secret_access_key:
            t += f"aws_secret_access_key = {self.aws.aws_secret_access_key}\n"
        if self.gcp.google_application_credentials.exists():
            t += f"gs_service_key_file = {self.gcp.google_application_credentials.as_posix()}\n"
        t += "\n[Boto]\n"
        t += "https_validate_certificates = True\n"
        t += "\n[GSUtil]\n"
        t += "content_language = en\n"
        t += "default_api_version = 2\n"
        if self.gcp.project: t += f"default_project_id = {self.gcp.project}\n"
        return t
    

    def write_botofile(
        self, 
        overwrite: bool = False, 
        **kwargs
    ):
        """
        Writes the boto config file to the path specified in the boto_config_path
        """
        if not self.aws.boto_config_exists or overwrite:
            logger.info(f"Writing boto config to {self.aws.boto_config_path.as_posix()}")
            self.aws.boto_config_path.write_text(self.get_boto_values())
    

    def update_config(self, **kwargs):
        """
        Update the config with the passed kwargs
        """
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            value = getattr(self, k)
            if isinstance(value, ConfigMixin):
                value.update_config(**v)
                continue
            if isinstance(value, pathlib.Path):
                value = pathlib.Path(v)
            setattr(self, k, value)

    def set_env(self):
        """
        Set the environment variables
        """
        self.aws.set_env()
        self.gcp.set_env()
        self.minio.set_env()
        self.s3c.set_env()
        self.r2.set_env()

    
    def update_auth(self, update_fs: bool = True, **config):
        """
        Updates the auth settings and optionally updates the fs
        """
        self.update_config(**config)
        self.set_env()


    def get_provider_config(self, provider: str, update_fs: bool = False, **auth_config) -> Tuple[Dict[str, Any], ProviderConfig]:
        """
        Updates the auth settings for the specified provider
        and returns the provider config
        """
        if not hasattr(self, provider): 
            raise ValueError(f"Provider {provider} not found")
        provider_spec: ConfigMixin = getattr(self, provider)
        if auth_config: provider_spec.update_auth(update_fs = update_fs, **auth_config)
        return provider_spec.build_fs_config(), provider_spec



from dataclasses import asdict, is_dataclass
from typing import Any, Optional, Type, Union

from temporalio import activity, workflow
from temporalio.worker import (
    ActivityInboundInterceptor,
    ExecuteActivityInput,
    ExecuteWorkflowInput,
    Interceptor,
    WorkflowInboundInterceptor,
    WorkflowInterceptorClassInput,
)

with workflow.unsafe.imports_passed_through():
    from sentry_sdk import Hub, capture_exception, set_context, set_tag


def _set_common_workflow_tags(info: Union[workflow.Info, activity.Info]):
    set_tag("temporal.workflow.type", info.workflow_type)
    set_tag("temporal.workflow.id", info.workflow_id)


class _SentryActivityInboundInterceptor(ActivityInboundInterceptor):
    async def execute_activity(self, input: ExecuteActivityInput) -> Any:
        # https://docs.sentry.io/platforms/python/troubleshooting/#addressing-concurrency-issues
        with Hub(Hub.current):
            set_tag("temporal.execution_type", "activity")
            set_tag("module", input.fn.__module__ + "." + input.fn.__qualname__)

            activity_info = activity.info()
            _set_common_workflow_tags(activity_info)
            set_tag("temporal.activity.id", activity_info.activity_id)
            set_tag("temporal.activity.type", activity_info.activity_type)
            set_tag("temporal.activity.task_queue", activity_info.task_queue)
            set_tag("temporal.workflow.namespace", activity_info.workflow_namespace)
            set_tag("temporal.workflow.run_id", activity_info.workflow_run_id)
            try:
                return await super().execute_activity(input)
            except Exception as e:
                if len(input.args) == 1 and is_dataclass(input.args[0]):
                    set_context("temporal.activity.input", asdict(input.args[0]))
                set_context("temporal.activity.info", activity.info().__dict__)
                capture_exception()
                raise e


class _SentryWorkflowInterceptor(WorkflowInboundInterceptor):
    async def execute_workflow(self, input: ExecuteWorkflowInput) -> Any:
        # https://docs.sentry.io/platforms/python/troubleshooting/#addressing-concurrency-issues
        with Hub(Hub.current):
            set_tag("temporal.execution_type", "workflow")
            set_tag("module", input.run_fn.__module__ + "." + input.run_fn.__qualname__)
            workflow_info = workflow.info()
            _set_common_workflow_tags(workflow_info)
            set_tag("temporal.workflow.task_queue", workflow_info.task_queue)
            set_tag("temporal.workflow.namespace", workflow_info.namespace)
            set_tag("temporal.workflow.run_id", workflow_info.run_id)
            try:
                return await super().execute_workflow(input)
            except Exception as e:
                if len(input.args) == 1 and is_dataclass(input.args[0]):
                    set_context("temporal.workflow.input", asdict(input.args[0]))
                set_context("temporal.workflow.info", workflow.info().__dict__)

                if not workflow.unsafe.is_replaying():
                    with workflow.unsafe.sandbox_unrestricted():
                        capture_exception()
                raise e


class SentryInterceptor(Interceptor):
    """Temporal Interceptor class which will report workflow & activity exceptions to Sentry"""

    def intercept_activity(
        self, next: ActivityInboundInterceptor
    ) -> ActivityInboundInterceptor:
        """Implementation of
        :py:meth:`temporalio.worker.Interceptor.intercept_activity`.
        """
        return _SentryActivityInboundInterceptor(super().intercept_activity(next))

    def workflow_interceptor_class(
        self, input: WorkflowInterceptorClassInput
    ) -> Optional[Type[WorkflowInboundInterceptor]]:
        _ = input
        return _SentryWorkflowInterceptor



from __future__ import annotations

import json
import pathlib
from typing import Optional, Dict, Any, List, Union

assets_path = pathlib.Path(__file__).parent
pricing_path = assets_path.joinpath('pricing')
presets_path = assets_path.joinpath('presets')

_provider_prices: Dict[str, List[Dict[str, Union[str, List, Dict, float]]]] = {}


def load_provider_prices(
    provider: Optional[str] = None,
    filepath: Optional[Union[str, pathlib.Path]] = None,
    overwrite: Optional[bool] = None,
    **kwargs,
) -> List[Dict[str, Union[str, List, Dict, float]]]:
    """
    Returns the provider prices
    """
    global _provider_prices
    if provider is None: provider = 'openai'
    elif filepath: provider = pathlib.Path(filepath).stem
    if provider not in _provider_prices or overwrite is True:
        if filepath is None: filepath = pricing_path.joinpath(f'{provider}.yaml')
        else: filepath = pathlib.Path(filepath)
        if not filepath.exists(): raise ValueError(f'{provider} Provider Pricing File does not exist: {filepath}')
        assert filepath.suffix in {
            ".yaml", ".yml", ".json"
        }, f"The {provider} Provider Pricing file must be a YAML or JSON file: {filepath}"
        if filepath.suffix == ".json":
            _provider_prices[provider] = json.loads(filepath.read_text())
        else:
            import yaml
            _provider_prices[provider] = yaml.safe_load(filepath.read_text())
    return _provider_prices[provider]


def load_preset_config(
    name: Optional[str] = None,
    filepath: Optional[Union[str, pathlib.Path]] = None,
    **overrides,
) -> Dict[str, Any]:
    """
    Returns the preset config
    """
    assert name or filepath, "You must provide either a name or a path to the preset"
    if filepath is None: filepath = presets_path.joinpath(f'{name}.yaml')
    else: filepath = pathlib.Path(filepath)
    if name is None: name = filepath.stem
    if not filepath.exists(): raise ValueError(f'{name} Preset Config File does not exist: {filepath}')
    assert filepath.suffix in {
        ".yaml", ".yml", ".json"
    }, f"The {name} preset file must be a YAML or JSON file: {filepath}"

    from lzo.utils import parse_envvars_from_text
    text = filepath.read_text()
    text, _ = parse_envvars_from_text(text)
    if filepath.suffix == ".json":
        data = json.loads(text)
    else:
        import yaml
        data = yaml.safe_load(text)
    if overrides: 
        from lzo.utils import update_dict
        data = update_dict(data, overrides)
    if data.get('models') and not isinstance(data['models'], list):
        if isinstance(data['models'], str):
            kws = {
                'filepath' if '/' in data['models'] else 'provider': data['models'],
            }
        elif isinstance(data['models'], dict):
            kws = data['models']
        else:
            raise ValueError(f'Invalid models: {data["models"]}')
        data['models'] = load_provider_prices(**kws)
    return data






from __future__ import annotations

"""
S3 Compatible Providers
"""

import typing as t
from ..cloudfs import CloudFileSystemMeta, BaseFileSystemAccessor

if t.TYPE_CHECKING:
    import s3fs
    import boto3
    import boto3.session
    from ..compat.r2 import R2FileSystem
    from s3transfer.manager import TransferManager
    from ...configs.providers import (
        AWSConfig,
        MinioConfig,
        S3CompatConfig,
        CloudflareR2Config,
    )


class AWSFileSystem(metaclass = CloudFileSystemMeta):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fsconfig: t.Optional['AWSConfig'] = None

    fs_name: str = 'aws'
    

class AWSAccessor(BaseFileSystemAccessor):
    """
    AWS File System Accessor
    """
    fsconfig: t.Optional['AWSConfig'] = None

    class CloudFileSystem(AWSFileSystem):
        pass


class MinioFileSystem(metaclass = CloudFileSystemMeta):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fsconfig: t.Optional['MinioConfig'] = None

    fs_name: str = 'minio'
    

class MinioAccessor(BaseFileSystemAccessor):
    """
    Minio File System Accessor
    """
    fsconfig: t.Optional['MinioConfig'] = None


    class CloudFileSystem(MinioFileSystem):
        pass


class S3CFileSystem(metaclass = CloudFileSystemMeta):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fsconfig: t.Optional['S3CompatConfig'] = None
    
    fs_name: str = 's3c'
    

class S3CAccessor(BaseFileSystemAccessor):
    """
    S3C File System Accessor
    """
    fsconfig: t.Optional['S3CompatConfig'] = None

    class CloudFileSystem(S3CFileSystem):
        pass

class R2FileSystem(metaclass = CloudFileSystemMeta):
    fs: 'R2FileSystem' = None
    fsa: 'R2FileSystem' = None
    fs_name: str = 'r2'
    config: t.Optional['CloudflareR2Config'] = None

class R2Accessor(BaseFileSystemAccessor):
    """
    R2 File System Accessor
    """
    fsconfig: t.Optional['CloudflareR2Config'] = None

    class CloudFileSystem(R2FileSystem):
        pass




from __future__ import annotations

""""
Cloud Filesystem
"""

import atexit
import functools
import typing as t
from lzl.require import LazyLib
from ..path import NormalAccessor
from ..utils.logs import logger
from .utils import (
    rewrite_async_syntax,
    create_method_fs,
    create_async_method_fs,
    create_staticmethod,
    create_async_coro,
)

if t.TYPE_CHECKING:
    from ..configs.main import FileIOConfig, ProviderConfig
    from fsspec.spec import AbstractFileSystem
    from fsspec.asyn import AsyncFileSystem
    from s3transfer.manager import TransferManager
    from boto3.session import Session

    _FST = t.TypeVar('_FST', bound=AbstractFileSystem)
    _ASFST = t.TypeVar('_ASFST', bound=AsyncFileSystem)

_MappedProviderAliases: t.Dict[str, str] = {
    's3fs': 'aws',
    'mc': 'minio',
}

_SupportedProviders: t.List[str] = [
    'aws',
    'gcp',
    'minio',
    's3c',
    'r2',
]

class CloudFileSystemMeta(type):
    """
    The Cloud File System
    """

    fs: t.Optional['_FST'] = None
    fsa: t.Optional['_ASFST'] = None
    fs_name: t.Optional[str] = None
    fsconfig: t.Optional['ProviderConfig'] = None

    boto: t.Optional['Session'] = None

    _s3t: t.Optional[t.Callable[..., 'TransferManager']] = None
    s3t: t.Optional['TransferManager'] = None
    # s3t: t.Optional[t.Union['TransferManager', t.Callable[..., 'TransferManager']]] = None
    
    _settings: t.Optional['FileIOConfig'] = None
    

    @property
    def settings(cls) -> 'FileIOConfig':
        """
        Returns the settings of the Cloud File System
        """
        if cls._settings is None:
            from lzl.io.file.utils import get_settings
            cls._settings = get_settings()
        return cls._settings
    
    def is_ready(cls) -> bool:
        """
        Returns whether the Cloud File System is ready
        """
        return bool(cls.fsa and cls.fs)
    

    def build_s3c(
        cls,
        provider: t.Optional[str] = 'aws',
        **auth_config: t.Any,
    ):
        """
        Builds the S3 Client
        """
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')

        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig

        fsspec_fs = s3fs.S3FileSystem
        if provider == 'r2':
            from .compat.r2 import R2FileSystem
            fsspec_fs = R2FileSystem

        config, pconfig = cls.settings.get_provider_config(provider, **auth_config)
        cls.fs = fsspec_fs(asynchronous = False, **config)
        cls.fsa = rewrite_async_syntax(fsspec_fs(asynchronous=True, **config))

        boto_config = BotoConfig(**pconfig.get_boto_config())
        cls.boto = boto3.client(
            's3',
            config = boto_config,
            **pconfig.get_boto_client_config()
        )
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = cls.settings.max_pool_connections
        )
        create_s3t = functools.partial(
            s3transfer.create_transfer_manager,
            cls.boto, 
            transfer_config
        )
        cls._s3t = create_s3t
        # cls.s3t = create_s3t()
        # if provider == 'r2':
        #     s3tm = create_s3t()
        #     cls.s3t = s3tm
        #     cls.fs.s3tm = s3tm
        #     cls.fsa.s3tm = s3tm
        # else:
        #     cls.s3t = property(create_s3t)
        cls.fsconfig = pconfig.model_copy()


    def build_filesystems(self, force: bool = False, **auth_config):
        """
        Lazily inits the filesystems
        """
        if self.fs is not None and self.fsa is not None and not force: 
            return
        
        if self.fs_name in _SupportedProviders:
            provider = _MappedProviderAliases.get(self.fs_name, self.fs_name)
            return self.build_s3c(provider = provider, **auth_config)
        raise NotImplementedError(f'Cloud File System `{self.fs_name}` is not supported')


    def reload_filesystem(cls):
        """ 
        Reinitializes the Filesystem
        """
        raise NotImplementedError



class BaseFileSystemAccessor(NormalAccessor):
    """
    Base File System Accessor
    """

    class CloudFileSystem(metaclass = CloudFileSystemMeta):
        pass


    info: t.Callable = create_staticmethod(CloudFileSystem, 'info')
    metadata: t.Callable = create_staticmethod(CloudFileSystem, 'metadata')
    stat: t.Callable = create_staticmethod(CloudFileSystem, 'stat')
    size: t.Callable = create_staticmethod(CloudFileSystem, 'size')
    exists: t.Callable = create_staticmethod(CloudFileSystem, 'exists')
    is_dir: t.Callable = create_staticmethod(CloudFileSystem, 'isdir')
    is_file: t.Callable = create_staticmethod(CloudFileSystem, 'isfile')
    copy: t.Callable = create_staticmethod(CloudFileSystem, 'copy')
    copy_file: t.Callable = create_staticmethod(CloudFileSystem, 'cp_file')
    get_file: t.Callable = create_staticmethod(CloudFileSystem, 'get_file')
    put_file: t.Callable = create_staticmethod(CloudFileSystem, 'put_file')
    metadata: t.Callable = create_staticmethod(CloudFileSystem, ['metadata', 'info'])
    checksum: t.Callable = create_staticmethod(CloudFileSystem, 'checksum')

    get: t.Callable = create_staticmethod(CloudFileSystem, 'get')
    put: t.Callable = create_staticmethod(CloudFileSystem, 'put')

    open: t.Callable = create_method_fs(CloudFileSystem, 'open')
    listdir: t.Callable = create_staticmethod(CloudFileSystem, 'ls')    
    walk: t.Callable = create_staticmethod(CloudFileSystem, 'walk')

    glob: t.Callable = create_method_fs(CloudFileSystem, 'glob')
    find: t.Callable = create_method_fs(CloudFileSystem, 'find')
    touch: t.Callable = create_method_fs(CloudFileSystem, 'touch')
    cat: t.Callable = create_method_fs(CloudFileSystem, 'cat')
    cat_file: t.Callable = create_method_fs(CloudFileSystem, 'cat_file')
    
    
    pipe: t.Callable = create_method_fs(CloudFileSystem, 'pipe')
    pipe_file: t.Callable = create_method_fs(CloudFileSystem, 'pipe_file')
    
    mkdir: t.Callable = create_method_fs(CloudFileSystem, 'mkdir')
    makedirs: t.Callable = create_method_fs(CloudFileSystem, ['makedirs', 'mkdirs'])
    unlink: t.Callable = create_method_fs(CloudFileSystem, 'rm_file')
    rmdir: t.Callable = create_method_fs(CloudFileSystem, 'rmdir')
    rename: t.Callable = create_method_fs(CloudFileSystem, 'rename')
    replace: t.Callable = create_method_fs(CloudFileSystem, 'rename')
    remove: t.Callable = create_method_fs(CloudFileSystem, 'rm')
    rm: t.Callable = create_staticmethod(CloudFileSystem, 'rm')
    rm_file: t.Callable = create_staticmethod(CloudFileSystem, 'rm_file')
    
    modified: t.Callable = create_method_fs(CloudFileSystem, 'modified')
    url: t.Callable = create_method_fs(CloudFileSystem, 'url')
    ukey: t.Callable = create_method_fs(CloudFileSystem, 'ukey')
    setxattr: t.Callable = create_method_fs(CloudFileSystem, 'setxattr')
    invalidate_cache: t.Callable = create_method_fs(CloudFileSystem, 'invalidate_cache')
    
    filesys: '_FST' = CloudFileSystem.fs
    afilesys: '_ASFST' = CloudFileSystem.fsa
    fsconfig: t.Optional['ProviderConfig'] = CloudFileSystem.fsconfig

    boto: t.Optional['Session'] = CloudFileSystem.boto
    _s3t: t.Optional[t.Callable[..., 'TransferManager']] = CloudFileSystem._s3t
    s3t: t.Optional['TransferManager'] = None 
    # CloudFileSystem.s3t
    # s3t: t.Optional[t.Union['TransferManager', t.Callable[..., 'TransferManager']]] = CloudFileSystem.s3t
    
    
    # Async Methods
    astat: t.Callable = create_async_coro(CloudFileSystem, 'stat')
    atouch: t.Callable = create_async_coro(CloudFileSystem, 'touch')
    aukey: t.Callable = create_async_coro(CloudFileSystem, 'ukey')
    asize: t.Callable = create_async_coro(CloudFileSystem, 'size')
    # aurl: t.Callable = create_async_coro(CloudFileSystem, 'url')
    # asetxattr: t.Callable = create_async_coro(CloudFileSystem, 'setxattr')
    aurl: t.Callable = create_async_method_fs(CloudFileSystem, 'url')
    asetxattr: t.Callable = create_async_method_fs(CloudFileSystem, 'setxattr')

    amodified: t.Callable = create_async_coro(CloudFileSystem, 'modified')
    ainvalidate_cache: t.Callable = create_async_coro(CloudFileSystem, 'invalidate_cache')
    arename: t.Callable = create_async_coro(CloudFileSystem, 'rename')
    areplace: t.Callable = create_async_coro(CloudFileSystem, 'rename')

    ainfo: t.Callable = create_async_method_fs(CloudFileSystem, 'ainfo')
    ametadata: t.Callable = create_async_method_fs(CloudFileSystem, 'ametadata')
    aexists: t.Callable = create_async_method_fs(CloudFileSystem, 'aexists')
    aglob: t.Callable = create_async_method_fs(CloudFileSystem, 'aglob')
    afind: t.Callable = create_async_method_fs(CloudFileSystem, 'afind')
    ais_dir: t.Callable = create_async_method_fs(CloudFileSystem, 'aisdir')
    ais_file: t.Callable = create_async_method_fs(CloudFileSystem, 'ais_file')
    acopy: t.Callable = create_async_method_fs(CloudFileSystem, 'acopy')
    acopy_file: t.Callable = create_async_method_fs(CloudFileSystem, 'acp_file')

    apipe: t.Callable = create_async_method_fs(CloudFileSystem, 'apipe')
    apipe_file: t.Callable = create_async_method_fs(CloudFileSystem, 'apipe_file')

    aget: t.Callable = create_async_coro(CloudFileSystem, 'aget')
    aget_file: t.Callable = create_async_coro(CloudFileSystem, 'aget_file')
    
    aput: t.Callable = create_async_method_fs(CloudFileSystem, 'aput')
    aput_file: t.Callable = create_async_method_fs(CloudFileSystem, 'aput_file')
    # ametadata: t.Callable = create_async_method_fs(CloudFileSystem, 'ametadata')
    aopen: t.Callable = create_async_method_fs(CloudFileSystem, '_open')
    amkdir: t.Callable = create_async_method_fs(CloudFileSystem, 'amkdir')
    amakedirs: t.Callable = create_async_method_fs(CloudFileSystem, 'amakedirs')
    aunlink: t.Callable = create_async_method_fs(CloudFileSystem, 'arm_file')
    armdir: t.Callable = create_async_method_fs(CloudFileSystem, 'armdir')
    aremove: t.Callable = create_async_method_fs(CloudFileSystem, 'arm')
    arm: t.Callable = create_async_method_fs(CloudFileSystem, 'arm')
    arm_file: t.Callable = create_async_coro(CloudFileSystem, 'arm_file')
    alistdir: t.Callable = create_async_method_fs(CloudFileSystem, ['alistdir', 'alist_objects'])
    awalk: t.Callable = create_async_method_fs(CloudFileSystem, 'awalk')

    is_fsspec: bool = True


    @classmethod
    def reload_cfs(cls, **kwargs):
        """
        Reloads the Cloud File System
        """
        cls.CloudFileSystem.build_filesystems(**kwargs)
        cls.info: t.Callable = create_staticmethod(cls.CloudFileSystem, 'info')
        cls.stat: t.Callable = create_staticmethod(cls.CloudFileSystem, 'stat')
        cls.size: t.Callable = create_staticmethod(cls.CloudFileSystem, 'size')
        cls.exists: t.Callable = create_staticmethod(cls.CloudFileSystem, 'exists')
        cls.is_dir: t.Callable = create_staticmethod(cls.CloudFileSystem, 'isdir')
        cls.is_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'isfile')
        cls.copy: t.Callable = create_staticmethod(cls.CloudFileSystem, 'copy')
        cls.copy_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'cp_file')
        cls.get_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'get_file')
        cls.put_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'put_file')
        cls.metadata: t.Callable = create_staticmethod(cls.CloudFileSystem, ['metadata', 'info'])

        cls.open: t.Callable = create_method_fs(cls.CloudFileSystem, 'open')
        cls.listdir: t.Callable = create_staticmethod(cls.CloudFileSystem, 'ls')    
        cls.walk: t.Callable = create_staticmethod(cls.CloudFileSystem, 'walk')
        cls.glob: t.Callable = create_staticmethod(cls.CloudFileSystem, 'glob')
        cls.get: t.Callable = create_staticmethod(cls.CloudFileSystem, 'get')
        cls.put: t.Callable = create_staticmethod(cls.CloudFileSystem, 'put')
        
        cls.checksum: t.Callable = create_method_fs(cls.CloudFileSystem, 'checksum')
        cls.cat: t.Callable = create_staticmethod(cls.CloudFileSystem, 'cat')
        cls.cat_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'cat_file')
        
        cls.pipe: t.Callable = create_staticmethod(cls.CloudFileSystem, 'pipe')
        cls.pipe_file: t.Callable = create_staticmethod(cls.CloudFileSystem, 'pipe_file')
    

        cls.find: t.Callable = create_method_fs(cls.CloudFileSystem, 'find')
        cls.touch: t.Callable = create_method_fs(cls.CloudFileSystem, 'touch')
        
        
        cls.mkdir: t.Callable = create_method_fs(cls.CloudFileSystem, 'mkdir')
        cls.makedirs: t.Callable = create_method_fs(cls.CloudFileSystem, ['makedirs', 'mkdirs'])
        cls.unlink: t.Callable = create_method_fs(cls.CloudFileSystem, 'rm_file')
        cls.rmdir: t.Callable = create_method_fs(cls.CloudFileSystem, 'rmdir')
        cls.rename : t.Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.replace : t.Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.rm : t.Callable = create_staticmethod(cls.CloudFileSystem, 'rm')
        cls.rm_file : t.Callable = create_staticmethod(cls.CloudFileSystem, 'rm_file')
        
        cls.remove : t.Callable = create_method_fs(cls.CloudFileSystem, 'rm')
        cls.modified: t.Callable = create_method_fs(cls.CloudFileSystem, 'modified')
        cls.setxattr: t.Callable = create_method_fs(cls.CloudFileSystem, 'setxattr')
        cls.url: t.Callable = create_method_fs(cls.CloudFileSystem, 'url')
        cls.ukey: t.Callable = create_method_fs(cls.CloudFileSystem, 'ukey')
        cls.invalidate_cache: t.Callable = create_method_fs(cls.CloudFileSystem, 'invalidate_cache')
        
        cls.filesys = cls.CloudFileSystem.fs
        cls.afilesys = cls.CloudFileSystem.fsa
        cls.fsconfig = cls.CloudFileSystem.fsconfig

        cls.boto = cls.CloudFileSystem.boto
        # cls.s3t = None #  cls.CloudFileSystem.s3t
        cls._s3t = cls.CloudFileSystem._s3t
        
        # Async Methods
        cls.astat: t.Callable = create_async_coro(cls.CloudFileSystem, 'stat')
        cls.atouch: t.Callable = create_async_coro(cls.CloudFileSystem, 'touch')
        cls.aukey: t.Callable = create_async_coro(cls.CloudFileSystem, 'ukey')
        
        # cls.aurl: t.Callable = create_async_coro(cls.CloudFileSystem, 'url')
        cls.asetxattr: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'setxattr')
        cls.aurl: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'url')
        # cls.asetxattr: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'setxattr')

        cls.amodified: t.Callable = create_async_coro(cls.CloudFileSystem, 'modified')
        cls.ainvalidate_cache: t.Callable = create_async_coro(cls.CloudFileSystem, 'invalidate_cache')
        cls.arename: t.Callable = create_async_coro(cls.CloudFileSystem, 'rename')
        cls.areplace: t.Callable = create_async_coro(cls.CloudFileSystem, 'rename')

        cls.ainfo: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'ainfo')
        cls.asize: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'size')
        cls.aexists: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aexists')

        cls.aglob: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aglob')
        cls.afind: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'afind')
        
        cls.acat: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'acat')
        cls.acat_file: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'acat_file')
        
        cls.apipe: t.Callable = create_async_coro(cls.CloudFileSystem, 'apipe')
        cls.apipe_file: t.Callable = create_async_coro(cls.CloudFileSystem, 'apipe_file')
        
        cls.ais_dir: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aisdir')
        cls.ais_file: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aisfile')
        cls.acopy: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'acopy')
        cls.acopy_file: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'acp_file')
        cls.aget: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aget')
        cls.aget_file: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aget_file')
        cls.aput: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aput')
        cls.aput_file: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'aput_file')
        cls.ametadata: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'ametadata')
        cls.aopen: t.Callable = create_async_method_fs(cls.CloudFileSystem, '_open')
        cls.amkdir: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'amkdir')
        cls.amakedirs: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'amakedirs')
        cls.aunlink: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'arm_file')
        cls.arm_file: t.Callable = create_async_coro(cls.CloudFileSystem, 'arm_file')
        cls.armdir: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'armdir')
        cls.aremove: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'arm')
        cls.arm: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'arm')
        cls.alistdir: t.Callable = create_async_method_fs(cls.CloudFileSystem, ['alistdir', 'alist_objects'])
        cls.awalk: t.Callable = create_async_method_fs(cls.CloudFileSystem, 'awalk')

    @classmethod
    def get_s3t(cls) -> 'TransferManager':
        """
        Returns the s3t transfer manager
        """
        if cls.s3t is None:
            cls.s3t = cls._s3t()
            atexit.register(cls._atexit_)
        return cls.s3t
    

    @classmethod
    def _atexit_(cls):
        """
        Cleans up the Cloud File System
        """
        if cls.s3t is not None:
            logger.info('Shutting Down S3 Transfer Manager')
            cls.s3t.shutdown()
            cls.s3t = None
    


from __future__ import annotations

import os
import yaml
import base64
from pathlib import Path
from lzo.types import BaseSettings, eproperty
from lzl.logging import logger, null_logger, Logger
from lzl.load import lazy_import
from pydantic import PrivateAttr, model_validator, field_validator, Field
from urllib.parse import urlparse, urlunparse, urljoin, parse_qsl, urlencode, quote
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Callable, Annotated, Generator

if TYPE_CHECKING:
    from fastapi import FastAPI
    from .manager import OAuth2Manager
    from .types.user import APIKeyUser
    from ..middlewares.persistent import PersistentMiddleware
    from .middleware import OAuth2Middleware
    from .backend import OAuth2Backend
    from lzl.io import JsonSerializer
    from lzl.io.persistence import TemporaryData
    from Cryptodome.Cipher._mode_ecb import EcbMode

here = Path(__file__).parent

_default_state_kws = {
    'token': '_token',
    'user': '_user',
    'apikey': '_apikey',
}

class OAuth2Config(BaseSettings):
    """
    The Primary OAuth2 Configuration
    """

    config_file: Optional[Path] = None
    config_dir: Optional[Path] = None

    header: Optional[str] = "Authorization"
    scheme: Optional[str] = "Bearer"

    domains: Optional[List[str]] = None
    debug_enabled: Optional[bool] = None

    # Session Configurations
    session_disabled: Optional[bool] = None
    if TYPE_CHECKING:
        session_class: Optional[Union[Type['PersistentMiddleware'], str]] = None
    else:
        session_class: Optional[Any] = None
    
    session_config: Optional[Dict[str, Any]] = None
    session_oauth2_key: Optional[str] = '_oauth2'
    session_state_kws: Optional[Dict[str, str]] = Field(default_factory = lambda: _default_state_kws)

    # Middleware Configurations
    if TYPE_CHECKING:
        manager_class: Optional[Union[Type['OAuth2Manager'], str]] = None
        middleware_class: Optional[Union[Type['OAuth2Middleware'], str]] = None
        backend_class: Optional[Union[Type['OAuth2Backend'], str]] = None
    else:
        manager_class: Optional[Any] = None
        middleware_class: Optional[Any] = None
        backend_class: Optional[Any] = None

    # Enrichment Configurations
    enable_user_info: Optional[bool] = None
    enable_query_params: Optional[bool] = None
    auth_query_param: Optional[str] = "token"
    api_key_query_param: Optional[str] = "apikey"

    # API Key Configurations
    api_keys_enabled: Optional[bool] = True
    allowed_api_keys: Optional[List[str]] = Field(default_factory = list)
    api_key_access_key: Optional[str] = None
    api_key_secret_key: Optional[str] = None

    # User Configurations
    staff_emails: Optional[List[str]] = Field(default_factory = list)
    admin_emails: Optional[List[str]] = Field(default_factory = list)
    sa_emails: Optional[List[str]] = Field(default_factory = list)

    staff_email_domains: Optional[List[str]] = Field(default_factory = list)
    admin_email_domains: Optional[List[str]] = Field(default_factory = list)
    sa_email_domains: Optional[List[str]] = Field(default_factory = list)

    disabled_paths: Optional[List[str]] = Field(default_factory = list)


    class Config:
        env_prefix = "OAUTH2_"
        arbitrary_types_allowed = True
        extra = 'allow'

    
    @property
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return logger if self.debug_enabled else null_logger


    @eproperty
    def manager(self) -> 'OAuth2Manager':
        """
        Returns the OAuth2 Manager
        """
        logger.info(f'Creating OAuth2 Manager: {self.manager_class}')
        return self.manager_class(config = self)
    
    @eproperty
    def aes_obj(self) -> Callable[..., 'EcbMode']:
        """
        Returns the AES Object
        """
        from lzl.ext.fast.utils.aes import create_aes_object
        return create_aes_object(self.api_key_secret_key, self.api_key_access_key)
    
    @eproperty
    def serializer(self) -> 'JsonSerializer':
        """
        Returns the serializer
        """
        from lzl.ext.fast.utils import serializer
        return serializer

    @eproperty
    def api_key_users(self) -> Dict[str, 'APIKeyUser']:
        """
        Returns the API Key Users
        """
        if 'api_key_users' not in self._extra:
            self._extra['api_key_users'] = self.parse_api_key_users(self.allowed_api_keys)
        return self._extra['api_key_users']

    def get_manager(self) -> 'OAuth2Manager':
        """
        Returns the OAuth2 Manager
        """
        # self._extra['mtg_class'] = mtg
        # logger.info(f'Creating OAuth2 Manager: {mtg}')
        return self.manager
    
    def parse_api_key_users(
        self, 
        api_keys: List[Union[str, Dict[str, Any]]],
    ) -> Dict[str, 'APIKeyUser']:
        """
        Parses the api key users
        """
        from .types.user import APIKeyUser
        api_key_users: Dict[str, 'APIKeyUser'] = {}
        for item in api_keys:
            api_key = APIKeyUser.parse_string(item) if isinstance(item, str) else APIKeyUser(**item)
            api_key_users[api_key.api_key] = api_key
        return api_key_users

    """
    Configuration Loaders
    """

    def load_domains(self, domains: Union[str, List[str], Dict[str, Union[str, List[str]]]]) -> None:
        """
        Loads the domains
        """
        if isinstance(domains, dict):
            domains = self.app_env.select(domains)
        if isinstance(domains, str): domains = [domains]
        if not self.domains: self.domains = []
        self.domains.extend(domains)
        self.domains = list(set(self.domains))


    def load_api_key_settings(self, data: Dict[str, Any]) -> None:
        """
        Loads the API Key Settings
        """
        if 'enabled' in data: self.api_keys_enabled = data['enabled']
        if data.get('access_key'): 
            self.api_key_access_key = data['access_key']
            assert len(self.api_key_access_key) == 16, "Invalid Access Key Length, must be 16 bytes"
        if data.get('secret_key'): 
            self.api_key_secret_key = data['secret_key']
            assert len(self.api_key_secret_key) in {32, 24, 16}, "Invalid Secret Key Length, must be 16, 24, or 32 bytes"
        if data.get('users'): self.api_key_users = self.parse_api_key_users(data['users'])
        logger.info(f'Loaded |g|{len(self.api_key_users)}|e| API Key Users', colored = True)

    def load_session_settings(self, data: Dict[str, Any]) -> None:
        """
        Loads the session settings
        """
        if 'disabled' in data: self.session_disabled = data['disabled']
        if data.get('config'): 
            self.session_config = data['config']
            if self.session_config.get('session_cookie') and '{env}' in self.session_config['session_cookie']:
                self.session_config['session_cookie'] = self.session_config['session_cookie'].replace('{env}', self.app_env.name)
            if self.session_config.get('state_cookie') and '{env}' in self.session_config['state_cookie']:
                self.session_config['state_cookie'] = self.session_config['state_cookie'].replace('{env}', self.app_env.name)
        if data.get('class'):
            self.session_class = lazy_import(data['class'])
        if data.get('state_kws'):
            self.session_state_kws.update(data['state_kws'])

    def load_middleware_settings(self, data: Union[str, Dict[str, Any]]) -> None:
        """
        Loads the middleware settings
        """
        if data.get('class'): self.middleware_class = lazy_import(data['class'])
        if data.get('backend'): self.backend_class = lazy_import(data['backend'])
        if 'user_info' in data and data['user_info'] is not None:
            self.enable_user_info = data['user_info']
        if 'query_params' in data and data['query_params'] is not None:
            query_data: Dict[str, Any] = data['query_params']
            self.enable_query_params = query_data.get('enabled', True)
            if query_data.get('auth'): self.auth_query_param = query_data['auth']
            if query_data.get('api_key'): self.api_key_query_param = query_data['api_key']

    def load_user_settings(self, users: Union[str, List[str], Dict[str, Union[str, List[str]]]]) -> None:
        """
        Loads the user settings
        """
        if users.get('admin_emails'):
            if not self.admin_emails: self.admin_emails = []
            self.admin_emails.extend(users['admin_emails'])
            self.admin_emails = list(set(self.admin_emails))
        if users.get('staff_emails'):
            if not self.staff_emails: self.staff_emails = []
            self.staff_emails.extend(users['staff_emails'])
            self.staff_emails = list(set(self.staff_emails))
        if users.get('sa_emails'):
            if not self.sa_emails: self.sa_emails = []
            self.sa_emails.extend(users['sa_emails'])
            self.sa_emails = list(set(self.sa_emails))
        if users.get('admin_email_domains'):
            if not self.admin_email_domains: self.admin_email_domains = []
            self.admin_email_domains.extend(users['admin_email_domains'])
            self.admin_email_domains = list(set(self.admin_email_domains))
        if users.get('staff_email_domains'):
            if not self.staff_email_domains: self.staff_email_domains = []
            self.staff_email_domains.extend(users['staff_email_domains'])
            self.staff_email_domains = list(set(self.staff_email_domains))
        if users.get('sa_email_domains'):
            if not self.sa_email_domains: self.sa_email_domains = []
            self.sa_email_domains.extend(users['sa_email_domains'])
            self.sa_email_domains = list(set(self.sa_email_domains))


    def load_from_config_file(self, config_file: Path) -> None:
        """
        Loads the config from a file
        """
        data: Dict[str, Union[Dict[str, Any], List[Dict[str, Any]], Any]] = yaml.safe_load(config_file.read_text())
        if data.get('manager'):
            self.manager_class = lazy_import(data['manager'])
        if data.get('session'): self.load_session_settings(data['session'])
        if data.get('api_keys'): self.load_api_key_settings(data['api_keys'])
        if data.get('users'): self.load_user_settings(data['users'])
        if data.get('domains'): self.load_domains(data['domains'])
        if data.get('providers'):
            self.manager.load_providers(data['providers'])
        if data.get('routes'): self.manager.update_routes(data['routes'])
        if data.get('middleware'): self.load_middleware_settings(data['middleware'])
        if 'debug_enabled' in data: self.debug_enabled = data['debug_enabled']
        # if data.get('routes'): self.manager.routes.update(data['routes'])


    """
    Helpers
    """

    def encrypt_api_key_data(self, data: Dict[str, Any]) -> str:
        """
        Encrypts the API Key Data
        """
        return base64.b64encode(self.aes_obj().encrypt(self.serializer.dumps(data).encode('ascii'))).decode()
    

    def decrypt_api_key_data(self, key: str) -> Dict[str, Any]:
        """
        Decrypts the API Key Data
        """
        return self.serializer.loads(self.aes_obj().decrypt(base64.b64decode(key)))

    
    @model_validator(mode = 'after')
    def validate_oauth_config(self):
        """
        Validates the config
        """
        if self.config_file:
            # Glob the config file
            if '*' in self.config_file.name:
                for config_file in self.config_file.parent.glob(self.config_file.name):
                    self.load_from_config_file(config_file)
            elif self.config_file.exists():
                self.load_from_config_file(self.config_file)
        
        if self.config_dir:
            # Glob the config dir for any yaml files
            for config_file in self.config_dir.glob('*.yaml'):
                self.load_from_config_file(config_file)

        if not self.session_disabled and not self.session_class:
            from ..middlewares.persistent import PersistentMiddleware
            self.session_class = PersistentMiddleware
        if not self.middleware_class:
            from .middleware import OAuth2Middleware
            self.middleware_class = OAuth2Middleware
        if not self.backend_class:
            from .backend import OAuth2Backend
            self.backend_class = OAuth2Backend
        
        if not self.manager_class:
            from .manager import OAuth2Manager
        elif isinstance(self.manager_class, str):
            self.manager_class = lazy_import(self.manager_class)
        return self
    
    @property
    def provider_pattern(self) -> str:
        """
        Returns the provider pattern

        '^(google|kinde|auth0)$'
        """
        if 'provider_pattern' not in self._extra:
            base_pattern = '|'.join(list(self.manager.providers.keys()))
            self._extra['provider_pattern'] = f'^({base_pattern})$'
        return self._extra['provider_pattern']
    
    @property
    def app(self) -> 'FastAPI':
        """
        Returns the app
        """
        return self._extra.get('app')
    
    @app.setter
    def app(self, value: 'FastAPI') -> None:
        """
        Sets the app
        """
        self._extra['app'] = value

    @property
    def temp_data(self) -> 'TemporaryData':
        """
        Returns the temporary data
        """
        if 'temp_data' not in self._extra:
            from lzl.io.persistence import TemporaryData
            filedir = os.getenv('OAUTH2_TEMP_DATA_DIR', here.joinpath('data'))
            self._extra['temp_data'] = TemporaryData(
                filedir = filedir,
                filepath = '_oauth2.json',
            )
        return self._extra['temp_data']


from __future__ import annotations

from .base import (
    is_primitive,
    serialize_object,
    deserialize_object,
    get_object_class,
    register_object_class,
    register_schema_mapping,
)

from .serialization import (
    ObjectDecoder,
    ObjectEncoder
)



from __future__ import annotations

"""Helpers for temporary, self-cleaning persistence files."""

import abc
import atexit
import contextlib
import os
import pathlib
import tempfile
import typing as t

from lzl import load
from lzl.logging import logger

if load.TYPE_CHECKING:
    import filelock
else:
    filelock = load.LazyLoad("filelock", install_missing=True)

if t.TYPE_CHECKING:
    from filelock import SoftFileLock

MutableMappingT = dict[str, t.Any]


class TemporaryData(abc.ABC):
    """Dictionary-like interface backed by an auto-cleaned JSON file."""

    def __init__(
        self,
        filepath: str | pathlib.Path | None = None,
        filedir: pathlib.Path | None = None,
        is_multithreaded: bool | None = False,
        timeout: int | None = 10,
    ) -> None:
        """Prepare the on-disk file and optional file lock used by the store."""

        if filedir is not None:
            if filepath is not None:
                filepath = filedir.joinpath(filepath)
            else:
                filepath = filedir.joinpath(f"{os.getpid()}.tmp.json")
        if not filepath:
            filepath = pathlib.Path(tempfile.mktemp())
        self.filepath = pathlib.Path(filepath)
        self._created_dir = not self.filepath.parent.exists()

        self.filepath.parent.mkdir(parents=True, exist_ok=True)
        self.filelock_path = self.filepath.with_suffix(".lock")
        self.timeout = timeout
        self.is_multithreaded = bool(is_multithreaded)
        self._filelock: "SoftFileLock | None" = None
        from lzl.io.ser import get_serializer

        self.serializer = get_serializer("json")

    @property
    def filelock(self) -> "SoftFileLock":
        """Return (and lazily create) the file lock guarding the JSON file."""

        if self._filelock is None:
            try:
                self._filelock = filelock.SoftFileLock(
                    self.filelock_path.as_posix(),
                    timeout=self.timeout,
                    thread_local=False,
                )
                with self._filelock.acquire():
                    if not self.filepath.exists():
                        self.filepath.write_text("{}")
                    data = self.serializer.loads(self.filepath.read_text())
                    if self.is_multithreaded and not data.get("process_id"):
                        data["process_id"] = os.getpid()
                        self.filepath.write_text(self.serializer.dumps(data, indent=2))
                atexit.register(self.cleanup_on_exit)
            except Exception as exc:  # pragma: no cover - logging path
                from lazyops.libs.logging import logger as lazy_logger

                lazy_logger.trace(f"Error creating filelock for {self.filepath}", exc)
                raise exc
        return self._filelock

    def _load_data(self) -> MutableMappingT:
        """Load the backing JSON document into memory."""

        if not self.filepath.exists():
            self.filepath.write_text("{}")
        return self.serializer.loads(self.filepath.read_text())

    @property
    def data(self) -> MutableMappingT:
        """Return the current JSON payload guarded by the file lock."""

        try:
            with self.filelock.acquire():
                return self._load_data()
        except filelock.Timeout as exc:  # pragma: no cover - timing dependent
            from lazyops.libs.logging import logger as lazy_logger

            lazy_logger.trace(f"Filelock timeout for {self.filepath}")
            raise exc

    @contextlib.contextmanager
    def ctx(self) -> t.Generator[MutableMappingT, None, None]:
        """Context manager yielding mutable JSON data with automatic flush."""

        try:
            with self.filelock.acquire():
                data = self._load_data()
                try:
                    yield data
                finally:
                    self.filepath.write_text(self.serializer.dumps(data, indent=2))
        except filelock.Timeout as exc:  # pragma: no cover - timing dependent
            logger.trace(f"Filelock timeout for {self.filepath}", exc)
            raise exc

    def get(self, key: str, default: t.Any | None = None) -> t.Any:
        """Return ``data[key]`` if present, otherwise ``default``."""

        return self.data.get(key, default)

    def __contains__(self, key: str) -> bool:
        """Return ``True`` when ``key`` exists in the cached data."""

        return key in self.data

    def __getitem__(self, key: str) -> t.Any:
        """Return the value stored for ``key`` or ``None`` when missing."""

        return self.data.get(key)

    def __setitem__(self, key: str, value: t.Any) -> None:
        """Persist ``value`` under ``key`` and flush to disk immediately."""

        with self.ctx() as data:
            data[key] = value

    def __delitem__(self, key: str) -> None:
        """Remove ``key`` from the stored data."""

        with self.ctx() as data:
            del data[key]

    def __iter__(self) -> t.Iterator[str]:
        """Return an iterator over stored keys."""

        return iter(self.data)

    def __len__(self) -> int:
        """Return the number of stored keys."""

        return len(self.data)

    def __repr__(self) -> str:  # pragma: no cover - convenience method
        return repr(self.data)

    def __str__(self) -> str:  # pragma: no cover - convenience method
        return str(self.data)

    def __bool__(self) -> bool:
        """Return ``True`` when the store contains at least one value."""

        return bool(self.data)

    def __eq__(self, other: t.Any) -> bool:
        """Compare the stored data with an arbitrary object."""

        return self.data == other

    def keys(self) -> t.Set[str]:
        """Return the set-like view of stored keys."""

        return self.data.keys()

    def setdefault(self, key: str, default: t.Any) -> t.Any:
        """Return ``data[key]`` if present, otherwise persist and return ``default``."""

        with self.ctx() as data:
            value = data.setdefault(key, default)
        return value

    def close(self) -> None:
        """Release the file lock safeguarding the JSON payload."""

        self.filelock.release()

    def append(self, key: str, value: t.Any) -> bool:
        """Append ``value`` to the list stored at ``key`` if it is unique."""

        with self.ctx() as data:
            if key not in data:
                data[key] = []
            if value not in data[key]:
                data[key].append(value)
                return False
        return True

    def cleanup_on_exit(self) -> None:
        """Remove the persisted files when the interpreter shuts down."""

        if not self.filepath.exists() and not self.filelock_path.exists():
            return
        if self.is_multithreaded and self["process_id"] != os.getpid():
            return
        with contextlib.suppress(Exception):
            self.close()
            self.filepath.unlink()
            self.filelock_path.unlink()
            if self._created_dir:
                contents = os.listdir(self.filepath.parent.as_posix())
                if not contents:
                    self.filepath.parent.rmdir()

    def has_logged(self, key: str) -> bool:
        """Return ``True`` if ``key`` has already been appended to ``logged``."""

        return self.append("logged", key)

    @classmethod
    def from_module(
        cls,
        module_name: str,
        data_dir: str | None = ".data",
        is_multithreaded: bool | None = False,
        **kwargs: t.Any,
    ) -> "TemporaryData":
        """Construct a :class:`TemporaryData` instance scoped to ``module_name``."""

        from lzo.utils.helpers.base import get_module_path

        module_path = get_module_path(module_name)
        module_dir = module_path.joinpath(data_dir or ".data")
        module_dir.mkdir(parents=True, exist_ok=True)
        filepath = module_dir.joinpath(f"{module_name}.tmp.json")
        if not filepath.exists():
            filepath.write_text("{}")
        return cls(filepath=filepath, is_multithreaded=is_multithreaded, **kwargs)


__all__ = ["TemporaryData"]



from __future__ import annotations

"""
MsgPack Serializer
"""

from typing import Any, Dict, Optional, Union, Type, TypeVar
from lzl.load import lazy_import
from .base import BinaryBaseSerializer, BaseModel, ModuleType, SchemaType, ObjectValue, logger
from .defaults import default_msgpack, MsgPackLibT, default_json, JsonLibT


MsgPackLibT = TypeVar("MsgPackLibT")

class MsgPackSerializer(BinaryBaseSerializer):
    name: Optional[str] = "msgpack"
    encoding: Optional[str] = "utf-8"
    disable_object_serialization: Optional[bool] = False
    jsonlib: JsonLibT = default_json
    msgpacklib: MsgPackLibT = default_msgpack

    def __init__(
        self, 
        msgpacklib: Optional[Union[str, Any]] = None,
        compression: Optional[str] = None,
        compression_level: int | None = None, 
        encoding: str | None = None, 
        serialization_obj: Optional[Type[BaseModel]] = None,
        serialization_obj_kwargs: Optional[Dict[str, Any]] = None,
        disable_object_serialization: Optional[bool] = None,
        jsonlib: Optional[Union[str, Any]] = None,
        **kwargs
    ):
        if not default_msgpack:
            raise ImportError("MsgPack Serializer is not available. Please install `msgpack`")
        super().__init__(compression = compression, compression_level = compression_level, encoding = encoding, **kwargs)
        self.serialization_obj = serialization_obj
        self.serialization_obj_kwargs = serialization_obj_kwargs or {}
        self.serialization_schemas: Dict[str, Type[BaseModel]] = {}
        if disable_object_serialization is not None:
            self.disable_object_serialization = disable_object_serialization
        
        if msgpacklib is not None:
            if isinstance(msgpacklib, str):
                msgpacklib = lazy_import(msgpacklib, is_module=True)
            assert hasattr(msgpacklib, "packb") and hasattr(msgpacklib, "unpackb"), f"Invalid MsgPack Library: {msgpacklib}"
            self.msgpacklib = msgpacklib
        self.msgpacklib_name = self.msgpacklib.__name__
        if jsonlib is not None:
            if isinstance(jsonlib, str):
                jsonlib = lazy_import(jsonlib, is_module=True)
            assert hasattr(jsonlib, "dumps") and hasattr(jsonlib, "loads"), f"Invalid JSON Library: {jsonlib}"
            self.jsonlib = jsonlib
        self.jsonlib_name = self.jsonlib.__name__
    

    @classmethod
    def set_default_lib(cls, lib: Union[str, MsgPackLibT, ModuleType]) -> None:
        """
        Sets the default MsgPack library
        """
        global default_msgpack
        if isinstance(lib, str):
            lib = lazy_import(lib, is_module=True)
        assert hasattr(lib, "packb") and hasattr(lib, "unpackb"), f"Invalid Msgpack Library: `{lib}`"
        cls.msgpacklib = lib
        default_msgpack = lib
        
    def default_serialization_hook(self, obj: ObjectValue):
        """
        Default Serialization Hook
        """
        if not isinstance(obj, BaseModel) and not hasattr(obj, 'model_dump'):
            logger.info(f'Invalid Object Type: |r|{type(obj)}|e| {obj}', colored = True, prefix = "msgpack")
            return obj
        
        if self.disable_object_serialization: 
            return obj.model_dump_json(**self.serialization_obj_kwargs)

        obj_class_name = self.fetch_object_classname(obj)
        if obj_class_name not in self.serialization_schemas:
            self.serialization_schemas[obj_class_name] = obj.__class__
        data = obj.model_dump(mode = 'json', **self.serialization_obj_kwargs)
        data['__class__'] = obj_class_name
        return self.msgpacklib.ExtType(2, self.jsonlib.dumps(data).encode(self.encoding))
    
    def default_deserialization_hook(self, code: int, data: Union[str, bytes]) -> ObjectValue:
        """
        Default Deserialization Hook
        """
        if code != 2: return data
        if isinstance(data, bytes): data = data.decode(self.encoding)
        try:
            data = self.jsonlib.loads(data)
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(data)}) {e}|e| {str(data)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
            return data
        if not self.disable_object_serialization:
            _class = data.pop('__class__', None)
            if _class is not None:
                if _class not in self.serialization_schemas:
                    self.serialization_schemas[_class] = lazy_import(_class)
                _class = self.serialization_schemas[_class]
                return _class.model_validate(data, **self.serialization_obj_kwargs)
        elif self.serialization_obj is not None:
            return self.serialization_obj.model_validate(data, **self.serialization_obj_kwargs)
        return data

    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> bytes:
        """
        Encode the value with the Pickle Library
        """
        if 'use_bin_type' not in kwargs: kwargs['use_bin_type'] = True
        if 'default' not in kwargs: kwargs['default'] = self.default_serialization_hook
        try:
            return self.msgpacklib.packb(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {str(value)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
        return None

    def decode_value(self, value: bytes, **kwargs) -> Union[SchemaType, Dict, Any]:
        """
        Decode the value with the Pickle Library
        """
        try:
            if 'raw' not in kwargs: kwargs['raw'] = False
            if 'ext_hook' not in kwargs: kwargs['ext_hook'] = self.default_deserialization_hook
            return self.msgpacklib.unpackb(value, **kwargs)
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {str(value)[:500]}', colored = True, prefix = "msgpack")
            if self.raise_errors: raise e
        return None


from __future__ import annotations

"""
Base DB Backend Manager
"""

import abc
import pathlib
import contextlib
from lzl.logging import logger, null_logger, Logger
from lzl.types import eproperty
from .config import DatabaseConfig, BaseDBSettings, BackendType, DBConfigT, DBSettingsT
from .utils import SQLAlchemyUtilities
import typing as t

if t.TYPE_CHECKING:
    from .backend import (
        BackendClassT,
        ThreadPool,
        Connection,
        AsyncConnection,
        SMSession,
        SMAsyncSession,
        Engine,
        AsyncEngine,
        ConnectionPool,
        SessionMakerT,
        AsyncSessionMakerT,
        
        SessionT,
        SessionGetT,
        AsyncSessionT,
        AsyncSessionGetT,
        SessionGetterT,
        ScopedSessionMakerT,
        AsyncScopedSessionMakerT,
    )
    from lzl.db.backends import DATABASE_BACKEND_CLASSES


DBManagerT = t.TypeVar('DBManagerT', bound = 'DBBackendManager')

class DBBackendManager(abc.ABC):
    """
    DB Backend Manager
    """    
    _extra: t.Dict[str, t.Any] = {}
    default_backend: t.Optional[BackendType] = 'sqlalchemy'
    db_type: t.Optional[str] = None

    def __init__(
        self,
        default_backend: t.Optional[BackendType] = None,
        debug_enabled: t.Optional[bool] = None,
        settings: t.Optional[DBSettingsT] = None,
        config: t.Optional[DBConfigT] = None,
        error_callbacks: t.Optional[t.List[t.Callable]] = None,
        # These exceptions will be handled by the error callbacks
        handled_exceptions: t.Optional[t.List[t.Type[Exception]]] = None,
        **kwargs,
    ):
        """
        Initializes the DB Backend Manager
        """
        if default_backend: self.default_backend = default_backend
        self.debug_enabled = debug_enabled
        self.current: t.Optional[str] = None
        self.backends: t.Dict[str, 'DATABASE_BACKEND_CLASSES'] = {}
        self._on_failure_callbacks: t.List[t.Callable] = error_callbacks or []
        self._handled_exceptions: t.List[t.Type[Exception]] = handled_exceptions or []
        self.db_type_name = self.db_type.capitalize() if self.db_type else 'DB'
        self.post_init(**kwargs)
        if not self.current and (settings or config):
            self.configure_backend(
                settings = settings,
                config = config,
                **kwargs,
            )
    
    def add_error_callback(self, callback: t.Callable):
        """
        Adds an error callback
        """
        self._on_failure_callbacks.append(callback)

    def post_init(self, **kwargs):
        """
        Post Initialization
        """
        pass

    def initialize_backend(
        self,
        name: t.Optional[str] = None,
        backend: t.Optional[BackendType] = None,

        settings: t.Optional[DBSettingsT] = None,
        config: t.Optional[DBConfigT] = None,
        
        filepath: t.Optional[t.Union[str, pathlib.Path]] = None, 
        env_var: t.Optional[str] = None,

        settings_app_name: t.Optional[str] = None,
        settings_filepath: t.Optional[t.Union[str, pathlib.Path]] = None,
        settings_env_var: t.Optional[str] = None,

        overrides: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> 'DATABASE_BACKEND_CLASSES':
        """
        Initializes the backend
        """
        if settings is None and (settings_filepath or settings_env_var):
            from lzl.db.configs import ADAPTER_TO_SETTINGS
            settings = ADAPTER_TO_SETTINGS[self.db_type](
                app_name = settings_app_name,
                env_var = settings_env_var,
                config_file = settings_filepath,
            )
        if settings and not config: 
            self.logger.info(f'Using Overrides: {overrides}', prefix = self.db_type_name, colored = True)
            config = settings.get_backend_config(
                filepath = filepath,
                env_var = env_var,
                **(overrides or {}),
            )
        assert config is not None, 'Config must be provided'
        name = name or config.name
        assert name not in self.backends, f'Backend {name} already exists'
        if backend is None: backend = config.backend or self.default_backend
        from lzl.db.backends import ADAPTER_TO_BACKENDS
        assert backend in ADAPTER_TO_BACKENDS[self.db_type], f'Invalid backend: {backend}'
        # assert backend in _BackendMapping, f'Invalid backend: {backend}'
        self.autologger.info(f'Initializing Backend |g|{name}|e| ({backend})', prefix = self.db_type_name, colored = True)
        backend_cls = ADAPTER_TO_BACKENDS[self.db_type][backend]
        # backend_cls = _BackendMapping[backend]
        return backend_cls(config = config, name = name, **kwargs)

    def configure_backend(
        self,
        name: t.Optional[str] = None,
        backend: t.Optional[BackendType] = None,
        settings: t.Optional[DBSettingsT] = None,
        config: t.Optional[DBConfigT] = None,
        
        filepath: t.Optional[t.Union[str, pathlib.Path]] = None, 
        env_var: t.Optional[str] = None,

        settings_app_name: t.Optional[str] = None,
        settings_filepath: t.Optional[t.Union[str, pathlib.Path]] = None,
        settings_env_var: t.Optional[str] = None,

        overrides: t.Optional[t.Dict[str, t.Any]] = None,
        set_as_current: t.Optional[bool] = None,
        **kwargs,
    ) -> 'DATABASE_BACKEND_CLASSES':
        """
        Configures the backend
        """
        bkend = self.initialize_backend(
            name = name,
            backend = backend,
            settings = settings,
            config = config,
            filepath = filepath,
            env_var = env_var,
            settings_app_name = settings_app_name,
            settings_filepath = settings_filepath,
            settings_env_var = settings_env_var,
            overrides = overrides,
            **kwargs,
        )
        self.backends[bkend.name] = bkend
        self.autologger.info(f'Configured Backend |g|{bkend.name}|e| ({bkend.backend})', prefix = 'PostgresDB', colored = True)
        if set_as_current or not self.current: 
            self.current = bkend.name
            self.autologger.info(f'Set Current Backend |g|{bkend.name}|e| ({bkend.backend})', prefix = 'PostgresDB', colored = True)
        return bkend

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return logger
    
    @eproperty
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return logger if self.debug_enabled else null_logger

    @eproperty
    def tpool(self) -> 'ThreadPool':
        """
        Returns the ThreadPool
        """
        from lzl.pool import ThreadPool
        return ThreadPool
    
    @eproperty
    def sql(self) -> 'SQLAlchemyUtilities':
        """
        Returns the SQLAlchemy Utilities
        """
        return SQLAlchemyUtilities()
    
    @property
    def bkend(self) -> 'DATABASE_BACKEND_CLASSES':
        """
        Returns the backend
        """
        return self.backends[self.current]


    def run_error_callbacks(
        self, 
        e: Exception,
        backend: 'BackendClassT',
        scoped: t.Optional[bool] = None,
        readonly: t.Optional[bool] = None,
    ):
        """
        Runs the error callbacks
        """
        import traceback
        tb = traceback.format_exc()
        partial_kws = {
            'error': e,
            'message': tb,
            'url': backend.url,
            'name': backend.name,
            'scoped': scoped,
            'readonly': readonly,
            'backend': backend.backend,
        }
        for cb in self._on_failure_callbacks:
            self.tpool.create_background_task(cb, **partial_kws)

    def _is_handled_exception(self, e: Exception) -> bool:
        # sourcery skip: use-any, use-next
        """
        Checks if the exception is handled
        """
        for cls in self._handled_exceptions:
            if isinstance(e, cls): return True
        return False
    
    @t.overload
    @contextlib.contextmanager
    def session(
        self, 
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.Generator['SMSession', None, None]:
        """
        Context Manager for the current session
        """
        ...

    @contextlib.contextmanager
    def session(
        self, 
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.Generator['SessionT', None, None]:
        """
        Context Manager for the current session
        """
        bkend = self.bkend
        ctx = ctx or self.current
        assert ctx is not None, 'Context must be set'
        assert ctx in self.backends, f'Invalid Context: {ctx} not in {list(self.backends.keys())}'
        sess_type, scoped = bkend._get_session(
            readonly = readonly,
            scoped = scoped,
            superuser = superuser,
            mode = 'sync',
        )
        sess: 'SessionT' = None
        try:
            sess: 'SessionT' = sess_type()
            yield sess
        except Exception as e:
            if not self._is_handled_exception(e):
                self.logger.trace(f'[{ctx} - {bkend.name}] Session error', e, depth = 5)
            self.run_error_callbacks(e, bkend, scoped, readonly)
            if auto_rollback is None: auto_rollback = bkend.config.auto_rollback_enabled
            if auto_rollback: 
                try:
                    sess.rollback()
                    self.logger.info('Session Rolled Back', prefix = f'{ctx} - {bkend.name}')
                except Exception as e:
                    self.logger.trace(f'[{ctx} - {bkend.name}] Session Rollback Error', e)
            if raise_errors: raise e
        finally:
            if auto_commit is None: auto_commit = bkend.config.autocommit_enabled
            if auto_commit:
                if scoped: sess_type.commit()
                else: sess.commit()
            if scoped: 
                sess_type.expire_all()
                sess_type.close()
                sess_type.remove()
            else:
                sess.close()

    @t.overload
    @contextlib.asynccontextmanager
    async def asession(
        self, 
        readonly: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.AsyncGenerator['SMAsyncSession', None]:
        ...

    @contextlib.asynccontextmanager
    async def asession(
        self, 
        readonly: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.AsyncGenerator['AsyncSessionGetT', None]:
        """
        Async Context Manager for the current session
        """
        if ctx is None: ctx = self.current
        assert ctx is not None, 'Context must be set'
        assert ctx in self.backends, f'Invalid Context: {ctx} not in {list(self.backends.keys())}'
        bkend = self.backends[ctx]
        sess_type, scoped = bkend._get_session(
            readonly = readonly,
            scoped = scoped,
            superuser = superuser,
            mode = 'async',
        )
        sess: 'AsyncSessionT' = None
        try:
            sess: 'AsyncSessionT' = sess_type()
            yield sess
        except Exception as e:
            if not self._is_handled_exception(e):
                self.logger.trace(f'[{ctx} - {bkend.name}] Session error', e, depth = 5)
            self.run_error_callbacks(e, bkend, scoped, readonly)
            if auto_rollback is None: auto_rollback = bkend.config.auto_rollback_enabled
            if auto_rollback: 
                try:
                    await sess.rollback()
                    self.logger.info('Session Rolled Back', prefix = f'{ctx} - {bkend.name}')
                except Exception as e:
                    self.logger.trace(f'[{ctx} - {bkend.name}] Session Rollback Error', e)
            if raise_errors: raise e
        finally:
            if auto_commit is None: auto_commit = bkend.config.autocommit_enabled
            if auto_commit:
                if scoped: await sess_type.commit()
                else: await sess.commit()
            if scoped: 
                sess_type.expire_all()
                await sess_type.close()
                await sess_type.remove()
            else:
                await sess.close()


    def database(
        self, 
        readonly: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        mode: t.Optional[t.Literal['sync', 'async']] = 'sync',
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.Callable[..., 'SessionGetT']:
        """
        Creates an inner dependency wrapper for the database session [FastAPI]
        """
        if mode == 'sync':
            def inner() -> t.Generator['SessionGetterT', None, None]:
                """
                Returns the database session
                """
                with self.session(readonly = readonly, auto_commit = auto_commit, raise_errors = raise_errors, auto_rollback = auto_rollback, superuser = superuser, ctx = ctx, **kwargs) as sess:
                    yield sess
        else:
            async def inner() -> t.AsyncGenerator['AsyncSessionGetT', None]:
                """
                Returns the database session
                """
                async with self.asession(readonly = readonly, auto_commit = auto_commit, raise_errors = raise_errors, auto_rollback = auto_rollback, superuser = superuser, ctx = ctx, **kwargs) as sess:
                    yield sess
        return inner

    def connection(
        self,
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        url: t.Optional[str] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.Generator['Connection', None, None]:
        """
        Context Manager for the connection
        """
        if ctx is None: ctx = self.current
        assert ctx is not None, 'Context must be set'
        assert ctx in self.backends, f'Invalid Context: {ctx} not in {list(self.backends.keys())}'
        bkend = self.backends[ctx]
        return bkend.connection(
            readonly = readonly,
            raise_errors = raise_errors,
            auto_rollback = auto_rollback,
            auto_commit = auto_commit,
            superuser = superuser,
            execution_options = execution_options,
            url = url,
            **kwargs,
        )
    
    def aconnection(
        self,
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        url: t.Optional[str] = None,
        ctx: t.Optional[str] = None,
        **kwargs,
    ) -> t.AsyncGenerator['AsyncConnection', None]:
        """
        Context Manager for the connection
        """
        if ctx is None: ctx = self.current
        assert ctx is not None, 'Context must be set'
        assert ctx in self.backends, f'Invalid Context: {ctx} not in {list(self.backends.keys())}'
        bkend = self.backends[ctx]
        return bkend.aconnection(
            readonly = readonly,
            raise_errors = raise_errors,
            auto_rollback = auto_rollback,
            auto_commit = auto_commit,
            superuser = superuser,
            execution_options = execution_options,
            url = url,
            **kwargs,
        )
    
    @t.overload
    @classmethod
    def default(
        cls: t.Type['DBManagerT'],
        name: t.Optional[str] = None,
        default_backend: t.Optional[BackendType] = None,
        debug_enabled: t.Optional[bool] = None,
        
        settings: t.Optional[DBSettingsT] = None,
        config: t.Optional[DBConfigT] = None,
        
        filepath: t.Optional[t.Union[str, pathlib.Path]] = None, 
        env_var: t.Optional[str] = None,

        settings_app_name: t.Optional[str] = None,
        settings_filepath: t.Optional[t.Union[str, pathlib.Path]] = None,
        settings_env_var: t.Optional[str] = None,

        overrides: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> 'DBManagerT':
        """
        Returns the default backend manager
        
        Args:
            name (t.Optional[str], optional): 
                The name of the backend. Defaults to None.
                This should be a unique name for the backend.

            default_backend (t.Optional[BackendType], optional): 
                The default backend type. Defaults to None.

            debug_enabled (t.Optional[bool], optional): 
                Whether to enable debug mode. Defaults to None.
            
            settings (t.Optional['DBSettingsT'], optional): 
                The settings object. Defaults to None.
            
            config (t.Optional['DBConfigT'], optional): 
                The config object. Defaults to None.
            
            filepath (t.Optional[t.Union[str, pathlib.Path]], optional): 
                The config file path. Defaults to None.
            
            env_var (t.Optional[str], optional): 
                The config env var. Defaults to None.
            
            settings_app_name (t.Optional[str], optional):
                The settings app name. Defaults to None.

            settings_filepath (t.Optional[t.Union[str, pathlib.Path]], optional):
                The settings config file path. Defaults to None.

            settings_env_var (t.Optional[str], optional):
                The settings config env var. Defaults to None.

            overrides (t.Optional[t.Dict[str, t.Any]], optional): 
                The config overrides. Defaults to None.
                This is used to override the config values.
        """
        ...


    @classmethod
    def default(
        cls: t.Type['DBManagerT'],
        name: t.Optional[str] = None,
        default_backend: t.Optional[BackendType] = None,
        debug_enabled: t.Optional[bool] = None,
        settings: t.Optional[DBSettingsT] = None,
        config: t.Optional[DBConfigT] = None,
        overrides: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> 'DBManagerT':
        """
        Returns the default backend manager
        """
        if not settings and not config:
            from lzl.db.configs import ADAPTER_TO_SETTINGS
            settings = ADAPTER_TO_SETTINGS[cls.db_type]()
        return cls(
            name = name,
            default_backend = default_backend,
            debug_enabled = debug_enabled,
            settings = settings,
            config = config,
            overrides = overrides,
            **kwargs,
        )
    
    def add_error_callback(self, *callbacks: t.Callable):
        """
        Adds an error callback
        """
        self._on_failure_callbacks.extend(callbacks)

    def add_handled_exception(self, *exceptions: t.Type[Exception]):
        """
        Adds an exception to the handled exceptions
        """
        self._handled_exceptions.extend(exceptions)
    
    def add_exception_callback(self, callback: t.Callable, *exceptions: t.Type[Exception]):
        """
        Adds an exception callback
        """
        self._on_failure_callbacks.append(callback)
        self._handled_exceptions.extend(exceptions)


    def __getattr__(self, name: str) -> t.Any:
        """
        Forward all unknown attributes to the backend
        """
        # logger.info(f'Forwarding Attribute: {name} to {self.current}', prefix = 'PGBackendManager', colored = True)
        if name in self.backends:
            return self.backends[name]
        return getattr(self.bkend, name)
    
    if t.TYPE_CHECKING:
        @property
        def engine(self) -> 'Engine':
            """
            Returns the engine
            """
            ...

        @property
        def engine_su(self) -> 'Engine':
            """
            Returns the superuser engine
            """

        @property
        def engine_ro(self) -> 'Engine':
            """
            Returns the readonly engine
            """

        @property
        def aengine(self) -> 'AsyncEngine':
            """
            [Async] Returns the engine
            """
            ...

        @property
        def aengine_su(self) -> 'AsyncEngine':
            """
            [Async] Returns the superuser engine
            """
        
        @property
        def aengine_ro(self) -> 'AsyncEngine':
            """
            [Async] Returns the readonly engine
            """

        @property
        def session_ro(self) -> 'SessionMakerT':
            """
            Returns the readonly session
            """
            ...

        @property
        def session_su(self) -> 'SessionMakerT':
            """
            Returns the superuser session
            """
        
        @property
        def asession_ro(self) -> 'AsyncSessionMakerT':
            """
            [Async] Returns the readonly session
            """
            ...

        @property
        def asession_su(self) -> 'AsyncSessionMakerT':
            """
            [Async] Returns the superuser session
            """
            ...

        @property
        def session_scoped(self) -> 'ScopedSessionMakerT':
            """
            Returns the Scoped Session 
            """
            ...
        
        @property
        def asession_scoped(self) -> 'AsyncScopedSessionMakerT':
            """
            [Async] Returns the Scoped Session
            """
            ...

            
        @property
        def pgpool(self) -> 'ConnectionPool':
            """
            Returns the Connection Pool
            """
        
        @property
        def pgpool_ro(self) -> 'ConnectionPool':
            """
            [Readonly] Returns the Connection Pool
            """
        
        @property
        def pgpool_su(self) -> 'ConnectionPool':
            """
            [Superuser] Returns the Connection Pool
            """
